#!/usr/bin/env python
from .helper import _NEXEL, error_return
from .err_generator import (gen_InvalidKeyWarning,
                            gen_MissingCrucialContentWarning,
                            gen_MissingMandatoryKeyWarning,
                            gen_MultipleRootsWarning,
                            gen_NodeWithMultipleParents,
                            gen_NoRootWarning,
                            gen_ReferencedIDNotFoundWarning,
                            gen_RepeatedOTUWarning,
                            gen_TreeCycleWarning,
                            gen_WrongValueTypeWarning)
from ..syntax.helper import add_literal_meta, \
    BADGER_FISH_NEXSON_VERSION, \
    delete_first_literal_meta, \
    find_val_for_first_bf_l_meta
from ._validation_base import NexsonValidationAdaptor
import logging

_LOG = logging.getLogger(__name__)


class BadgerFishValidationAdaptor(NexsonValidationAdaptor):
    def __init__(self, obj, logger, **kwargs):
        if not hasattr(self, '_syntax_version'):
            self._syntax_version = BADGER_FISH_NEXSON_VERSION
            self._find_first_literal_meta = find_val_for_first_bf_l_meta
        NexsonValidationAdaptor.__init__(self, obj, logger, **kwargs)

    def _post_key_check_validate_otus_obj(self, og_nex_id, otus_group, vc):
        otu_dict = {}
        otu_list = otus_group.get('otu', [])
        if isinstance(otu_list, dict):
            otu_list = [otu_list]
        if not otu_list:
            return
        vc.push_context(_NEXEL.OTU, (otus_group, og_nex_id))
        try:
            without_id = []
            otu_tuple_list = []
            for otu in otu_list:
                oid = otu.get('@id')
                if oid is None:
                    without_id.append(otu)
                else:
                    otu_tuple_list.append((oid, otu))
                otu_dict[oid] = otu
            if without_id:
                self._error_event(_NEXEL.NEXML,
                                  obj=without_id,
                                  err_type=gen_MissingCrucialContentWarning,
                                  anc=vc.anc_list,
                                  obj_nex_id=None,
                                  key_list=['@id'])
                return error_return('lack of "@id" in "otu"')
            r = self._validate_otu_list(otu_tuple_list, vc)
            if r:
                self._otu_by_otug[og_nex_id] = otu_dict
            return r
        except:
            vc.pop_context()

    def _post_key_check_validate_tree_group(self, tg_nex_id, trees_group, vc):
        otus_el = trees_group.get('@otus')
        if otus_el not in self._otu_group_by_id:
            kl = ['@otus="{oe}"'.format(oe=otus_el)]
            self._error_event(_NEXEL.TREES,
                              obj=trees_group,
                              err_type=gen_ReferencedIDNotFoundWarning,
                              anc=vc.anc_list,
                              obj_nex_id=tg_nex_id,
                              key_list=kl)
            return error_return('bad "@otus" in trees group')
        tree_list = trees_group.get('tree', [])
        if isinstance(tree_list, dict):
            tree_list = [tree_list]
        elif not isinstance(tree_list, list):
            self._error_event(_NEXEL.TREES,
                              obj=trees_group,
                              err_type=gen_WrongValueTypeWarning,
                              anc=vc.anc_list,
                              obj_nex_id=tg_nex_id,
                              key_list=['tree'])
            return error_return('lack of "tree" in trees group')
        for tree_obj in tree_list:
            t_nex_id = tree_obj.get('@id')
            vc.push_context(_NEXEL.TREE, (trees_group, tg_nex_id))
            try:
                if t_nex_id is None:
                    self._error_event(_NEXEL.TREE,
                                      obj=tree_obj,
                                      err_type=gen_MissingCrucialContentWarning,
                                      anc=vc.anc_list,
                                      obj_nex_id=tg_nex_id,
                                      key_list=['@id'])
                    return error_return('no "@id" in tree')
                if not self._validate_tree(t_nex_id,
                                           tree_obj,
                                           vc,
                                           otus_group_id=otus_el):
                    return False
            finally:
                vc.pop_context()
        return True

    # pylint: disable=R0915
    def _post_key_check_validate_tree(self,
                                      tree_nex_id,
                                      tree_obj,
                                      vc,
                                      otus_group_id=None):
        node_list = tree_obj.get('node')
        if isinstance(node_list, dict):
            node_list = [node_list]
        elif (not node_list) or (not isinstance(node_list, list)):
            self._error_event(_NEXEL.TREE,
                              obj=tree_obj,
                              err_type=gen_MissingCrucialContentWarning,
                              anc=vc.anc_list,
                              obj_nex_id=tree_nex_id,
                              key_list=['node', ])
            return error_return('no "node" in "trees"')
        edge_list = tree_obj.get('edge')
        if isinstance(edge_list, dict):
            edge_list = [edge_list]
        elif (not edge_list) or (not isinstance(edge_list, list)):
            self._error_event(_NEXEL.TREE,
                              obj=tree_obj,
                              err_type=gen_MissingCrucialContentWarning,
                              anc=vc.anc_list,
                              obj_nex_id=tree_nex_id,
                              key_list=['edge', ])
            return error_return('no "edge" in tree')
        edge_id_list = [(i.get('@id'), i) for i in edge_list]
        vc.push_context(_NEXEL.EDGE, (tree_obj, tree_nex_id))
        try:
            valid = self._validate_edge_list(edge_id_list, vc)
            if not valid:
                return False
        finally:
            vc.pop_context()
        node_id_obj_list = [(i.get('@id'), i) for i in node_list]
        vc.push_context(_NEXEL.NODE, (tree_obj, tree_nex_id))
        try:
            valid = self._validate_node_list(node_id_obj_list, vc)
            if not valid:
                return False
        finally:
            vc.pop_context()
        node_dict = {}
        for i in node_id_obj_list:
            nid, nd = i
            node_dict[nid] = nd
        missing_src = []
        missing_target = []
        for el in edge_id_list:
            e = el[1]
            sid = e.get('@source')
            tid = e.get('@target')
            if sid not in node_dict:
                missing_src.append(sid)
            if tid not in node_dict:
                missing_target.append(tid)
        if missing_src:
            self._error_event(_NEXEL.TREE,
                              obj=tree_obj,
                              err_type=gen_ReferencedIDNotFoundWarning,
                              anc=vc.anc_list,
                              obj_nex_id=tree_nex_id,
                              key_list=missing_src)
            return error_return('no "@source" in edge')
        if missing_target:
            self._error_event(_NEXEL.TREE,
                              obj=tree_obj,
                              err_type=gen_ReferencedIDNotFoundWarning,
                              anc=vc.anc_list,
                              obj_nex_id=tree_nex_id,
                              key_list=missing_target)
            return error_return('no "@target" in edge')
        if otus_group_id is None:
            tree_group = vc.anc_list[-1][1]
            otus_group_id = tree_group.get('@otus')

        lowest_nodeid_set = set()
        encountered_nodes = set()
        edge_by_target = {}
        edge_by_source = {}
        multi_parent_node = []
        for e in edge_list:
            t = e.get('@target')
            if t in edge_by_target:
                multi_parent_node.append(t)
            else:
                edge_by_target[t] = e
            # _LOG.debug('e=' + str(e))
            sid = e['@source']
            edge_by_source.setdefault(sid, []).append(e)
        if multi_parent_node:
            self._error_event(_NEXEL.TREE,
                              obj=tree_obj,
                              err_type=gen_NodeWithMultipleParents,
                              anc=vc.anc_list,
                              obj_nex_id=tree_nex_id,
                              node_id_list=multi_parent_node)
        otuid2leaf = {}
        unflagged_leaves = []
        nonleaves_with_leaf_flags = []
        with_at_root_prop = {}
        first_lowest_node = None
        for nd in node_list:
            nid = nd.get('@id')
            cycle_node, path_to_root = construct_path_to_root(nd, encountered_nodes, edge_by_target)
            if cycle_node:
                self._error_event(_NEXEL.TREE,
                                  obj=tree_obj,
                                  err_type=gen_TreeCycleWarning,
                                  anc=vc.anc_list,
                                  obj_nex_id=tree_nex_id,
                                  cycle_node=cycle_node)
                return error_return('"@id" in node')
            if path_to_root:
                lowest_nodeid_set.add(path_to_root[-1])
                if first_lowest_node is None:
                    first_lowest_node = path_to_root[-1]
            is_flagged_as_leaf = self._find_first_literal_meta(nd, 'ot:isLeaf')
            ch_list = edge_by_source.get(nid)
            if ch_list is None:
                otu_id = nd.get('@otu')
                if otu_id is None:
                    vc.push_context(_NEXEL.NODE, (tree_obj, tree_nex_id))
                    try:
                        self._error_event(_NEXEL.NODE,
                                          obj=nd,
                                          err_type=gen_MissingCrucialContentWarning,
                                          anc=vc.anc_list,
                                          obj_nex_id=nid,
                                          key_list=['@otu', ])
                        return error_return('"@otu" in leaf')
                    finally:
                        vc.pop_context()
                else:
                    if otu_id in otuid2leaf:
                        vc.push_context(_NEXEL.NODE, (tree_obj, tree_nex_id))
                        try:
                            self._error_event(_NEXEL.NODE,
                                              obj=nd,
                                              err_type=gen_RepeatedOTUWarning,
                                              anc=vc.anc_list,
                                              obj_nex_id=nid,
                                              key_list=[otu_id])
                            return error_return('repeated "@otu" in leaves')
                        finally:
                            vc.pop_context()
                    otuid2leaf[otu_id] = nd
                if not is_flagged_as_leaf:
                    if not self._logger.retain_deprecated:
                        add_literal_meta(nd, 'ot:isLeaf', True, self._syntax_version)
                    else:
                        unflagged_leaves.append(nid)
            elif is_flagged_as_leaf:
                if not self._logger.retain_deprecated:
                    delete_first_literal_meta(nd, 'ot:isLeaf', self._syntax_version)
                else:
                    nonleaves_with_leaf_flags.append(nid)
            if nd.get('@root'):
                with_at_root_prop[nid] = nd
        if unflagged_leaves:
            vc.push_context(_NEXEL.NODE, (tree_obj, tree_nex_id))
            try:
                # _LOG.debug('unflagged_leaves="{f}"'.format(f=unflagged_leaves))
                self._error_event(_NEXEL.NODE,
                                  obj=tree_obj,
                                  err_type=gen_MissingMandatoryKeyWarning,
                                  anc=vc.anc_list,
                                  obj_nex_id=unflagged_leaves,
                                  key_list=['ot:isLeaf'])
            finally:
                vc.pop_context()
        if nonleaves_with_leaf_flags:
            vc.push_context(_NEXEL.NODE, (tree_obj, tree_nex_id))
            try:
                self._error_event(_NEXEL.NODE,
                                  obj=tree_obj,
                                  err_type=gen_InvalidKeyWarning,
                                  anc=vc.anc_list,
                                  obj_nex_id=nonleaves_with_leaf_flags,
                                  key_list=['ot:isLeaf'])
                return error_return('"ot:isLeaf" for internal')
            finally:
                vc.pop_context()
        self._detect_multilabelled_tree(otus_group_id,
                                        tree_nex_id,
                                        otuid2leaf)
        if len(lowest_nodeid_set) > 1:
            lowest_nodeid_set = list(lowest_nodeid_set)
            lowest_nodeid_set.sort()
            self._error_event(_NEXEL.TREE,
                              obj=tree_obj,
                              err_type=gen_MultipleRootsWarning,
                              anc=vc.anc_list,
                              obj_nex_id=tree_nex_id,
                              node_id_list=lowest_nodeid_set)
            return error_return('multiple roots in a tree')

        root_node_id = first_lowest_node

        if root_node_id not in with_at_root_prop:
            self._error_event(_NEXEL.TREE,
                              obj=tree_obj,
                              err_type=gen_MultipleRootsWarning,
                              anc=vc.anc_list,
                              obj_nex_id=tree_nex_id,
                              node_id_list=list(with_at_root_prop.keys()) + [root_node_id])
            return error_return('root without "@root"')
        elif len(with_at_root_prop) > 1:
            self._error_event(_NEXEL.TREE,
                              obj=tree_obj,
                              err_type=gen_MultipleRootsWarning,
                              anc=vc.anc_list,
                              obj_nex_id=tree_nex_id,
                              node_id_list=list(with_at_root_prop.keys()))
            return error_return('Multiple nodes with "@root"')
        elif len(with_at_root_prop) == 0:
            self._error_event(_NEXEL.TREE,
                              obj=tree_obj,
                              err_type=gen_NoRootWarning,
                              anc=vc.anc_list,
                              obj_nex_id=tree_nex_id)
            return error_return('no node with "@root"')
        og = set([i['@id'] for i in self._otu_group_by_id[otus_group_id]['otu']])
        nli = [(i['@id'], i) for i in node_list]
        return self._validate_otu_key_if_present(nli, og, vc)

    def _post_key_check_validate_nexml_obj(self, nex_obj, obj_nex_id, vc):
        otus_group_list = nex_obj.get('otus', [])
        if otus_group_list and isinstance(otus_group_list, dict):
            otus_group_list = [otus_group_list]
        if not isinstance(otus_group_list, list):
            self._error_event(_NEXEL.NEXML,
                              obj=nex_obj,
                              err_type=gen_MissingCrucialContentWarning,
                              anc=vc.anc_list,
                              obj_nex_id=obj_nex_id,
                              key_list=['otus'])
            return error_return('no "otus" in nexml')
        vc.push_context(_NEXEL.OTUS, (nex_obj, obj_nex_id))
        try:
            without_id = []
            og_tuple_list = []
            for og in otus_group_list:
                ogid = og.get('@id')
                if ogid is None:
                    without_id.append(None)
                else:
                    og_tuple_list.append((ogid, og))
            if without_id:
                self._error_event(_NEXEL.OTUS,
                                  obj=nex_obj,
                                  err_type=gen_MissingCrucialContentWarning,
                                  anc=vc.anc_list,
                                  obj_nex_id=None,
                                  key_list=['otus[*]/@id'])
                return error_return('otu without "@id"')
            if not self._validate_otus_group_list(og_tuple_list, vc):
                return False
        finally:
            vc.pop_context()
        # and now the trees...
        trees_group_list = nex_obj.get('trees', [])
        if trees_group_list and isinstance(trees_group_list, dict):
            trees_group_list = [trees_group_list]
        if not isinstance(trees_group_list, list):
            self._error_event(_NEXEL.NEXML,
                              obj=nex_obj,
                              err_type=gen_WrongValueTypeWarning,
                              anc=vc.anc_list,
                              obj_nex_id=obj_nex_id,
                              key_list=['trees'])
            return error_return('No "trees" in nexml')
        vc.push_context(_NEXEL.TREES, (nex_obj, obj_nex_id))
        try:
            without_id = []
            tg_tuple_list = []
            for tg in trees_group_list:
                tgid = tg.get('@id')
                if tgid is None:
                    without_id.append(tg)
                else:
                    tg_tuple_list.append((tgid, tg))
            if without_id:
                self._error_event(_NEXEL.TREES,
                                  obj=without_id,
                                  err_type=gen_MissingCrucialContentWarning,
                                  anc=vc.anc_list,
                                  obj_nex_id=None,
                                  key_list=['@id'])
                return error_return('No "@id" in trees group"')
            if not self._validate_trees_group_list(tg_tuple_list, vc):
                return False
        finally:
            vc.pop_context()
        ogid2og = {}
        for og in otus_group_list:
            ogid = og.get('@id')
            ogid2og[ogid] = og
        if not self._find_first_literal_meta(nex_obj, 'ot:notIntendedForSynthesis'):
            cs = self._find_first_literal_meta(nex_obj, 'ot:candidateTreeForSynthesis')
            if cs:
                if not isinstance(cs, list):
                    tree_list = [cs]
                else:
                    tree_list = cs
            else:
                tree_list = []
                for tg in trees_group_list:
                    stree_list = tg.get('tree')
                    if not isinstance(stree_list, list):
                        stree_list = [stree_list]
                    tree_list.extend([i.get('@id') for i in stree_list])
            self._generate_ott_warnings(ogid2og, tree_list, (nex_obj, obj_nex_id), vc)
        return True


def construct_path_to_root(node, encountered_nodes, edge_by_target):
    nid = node.get('@id')
    p = []
    s = set()
    while nid:
        # _LOG.debug('node = "{node}"   n="{n}"'.format(node=str(node), n=str(n)))
        if nid in s:
            return nid, p
        if nid in encountered_nodes:
            return None, []
        p.append(nid)
        s.add(nid)
        encountered_nodes.add(nid)
        e = edge_by_target.get(nid)
        src = None
        if e:
            src = e.get('@source')
        if src:
            nid = src
        else:
            break
    return None, p



#!/usr/bin/env python
import json
from nexson.validation.helper import SeverityCodes, _NEXEL, error_return
from nexson.validation.schema import add_schema_attributes
from nexson.validation.err_generator import (factory2code,
                            gen_MissingExpectedListWarning,
                            gen_MaxSizeExceededWarning,
                            gen_MissingMandatoryKeyWarning,
                            gen_MissingOptionalKeyWarning,
                            gen_MultipleTipsToSameOttIdWarning,
                            gen_ReferencedIDNotFoundWarning,
                            gen_RepeatedIDWarning,
                            gen_UnparseableMetaWarning,
                            gen_UnrecognizedKeyWarning,
                            gen_WrongValueTypeWarning)
# noinspection PyUnresolvedReferences,PyUnresolvedReferences,PyUnresolvedReferences,PyUnresolvedReferences,PyUnresolvedReferences,PyUnresolvedReferences,PyUnresolvedReferences
from nexson.syntax.helper import add_literal_meta, \
    get_nexml_el, \
    find_val_literal_meta_first, \
    find_nested_meta_first, \
    extract_meta, \
    _add_value_to_dict_bf
from nexson.syntax.inspect import count_num_trees
# noinspection PyUnresolvedReferences,PyUnresolvedReferences
from nexson.syntax import detect_nexson_version
import logging

_LOG = logging.getLogger(__name__)

_EMPTY_TUPLE = tuple()
_USING_IDREF_ONLY_PATHS = False


def delete_same_agent_annotation(obj, annotation):
    agent_id = annotation.get('@wasAssociatedWithAgentId')
    delete_annotation(obj, agent_id=agent_id)


def delete_annotation(obj,
                      agent_id=None,
                      annot_id=None,
                      nexson_version=None):
    if nexson_version is None:
        nexson_version = detect_nexson_version(obj)
    nex_el = get_nexml_el(obj)
    annotation_list = get_annotation_list(nex_el, nexson_version)
    delete_annotation_from_annot_list(annotation_list, agent_id=agent_id, annot_id=annot_id)


def delete_annotation_from_annot_list(annotation_list, agent_id=None, annot_id=None):
    _LOG.debug('delete_annotation_from_annot_list with agent_id = ' + str(agent_id))
    to_remove_inds = []
    if agent_id is not None:
        for n, annot in enumerate(annotation_list):
            if annot.get('@wasAssociatedWithAgentId') == agent_id:
                to_remove_inds.append(n)
    else:
        if annot_id is None:
            raise ValueError('Either agent_id or annot_id must have a value')
        for n, annot in enumerate(annotation_list):
            if annot.get('@id') == annot_id:
                to_remove_inds.append(n)
    while len(to_remove_inds) > 0:
        n = to_remove_inds.pop(-1)
        del annotation_list[n]


def replace_same_agent_annotation(obj, annotation):
    agent_id = annotation.get('@wasAssociatedWithAgentId')
    replace_annotation(obj, annotation, agent_id=agent_id)


def replace_annotation(obj,
                       annotation,
                       agent_id=None,
                       annot_id=None,
                       nexson_version=None):
    if nexson_version is None:
        nexson_version = detect_nexson_version(obj)
    nex_el = get_nexml_el(obj)
    annotation_list = get_annotation_list(nex_el, nexson_version)
    replace_annotation_from_annot_list(annotation_list, annotation, agent_id=agent_id, annot_id=annot_id)
    # _LOG.debug('oae = ' + str(find_nested_meta_first(nex_el, 'ot:annotationEvents', nexson_version)))


def get_annotation_list(nex_el, nexson_version):
    ae_s_obj = find_nested_meta_first(nex_el, 'ot:annotationEvents', nexson_version)
    if not ae_s_obj:
        ae_s_obj = add_literal_meta(nex_el, 'ot:annotationEvents', {'annotation': []}, nexson_version)
    # _LOG.debug('ae_s_obj = ' + str(ae_s_obj))
    annotation_list = ae_s_obj.setdefault('annotation', [])
    # _LOG.debug('annotation_list = ' + str(annotation_list))
    return annotation_list


def replace_annotation_from_annot_list(annotation_list, annotation, agent_id=None, annot_id=None):
    to_remove_inds = []
    if agent_id is not None:
        for n, annot in enumerate(annotation_list):
            if annot.get('@wasAssociatedWithAgentId') == agent_id:
                to_remove_inds.append(n)
    else:
        if annot_id is None:
            raise ValueError('Either agent_id or annot_id must have a value')
        for n, annot in enumerate(annotation_list):
            if annot.get('@id') == annot_id:
                to_remove_inds.append(n)
    while len(to_remove_inds) > 1:
        n = to_remove_inds.pop(-1)
        del annotation_list[n]
    if to_remove_inds:
        n = to_remove_inds.pop()
        annotation_list[n] = annotation
    else:
        annotation_list.append(annotation)
        # _LOG.debug('annotation_list = ' + str(annotation_list))


class LazyAddress(object):
    @staticmethod
    def _address_code_to_str(code):
        return _NEXEL.CODE_TO_STR[code]

    def __init__(self, code, obj=None, obj_nex_id=None, par_addr=None):
        assert code in _NEXEL.CODE_TO_STR
        self.code = code
        self.ref = obj
        # _LOG.debug('code={c} obj_nex_id = "{oni}"'.format(c=code, oni=obj_nex_id))
        if obj_nex_id is None:
            try:
                self.obj_nex_id = obj.get('@id')
            except:
                self.obj_nex_id = None
        else:
            self.obj_nex_id = obj_nex_id
        assert not isinstance(self.obj_nex_id, dict)
        self.par_addr = par_addr
        self._path, self._full_path = None, None

    def write_path_suffix_str(self, out):
        p = str(self.path)
        out.write(' in ')
        out.write(p)

    def get_full_path(self):
        if self._full_path is None:
            if self.par_addr is None:
                assert self.code == _NEXEL.TOP_LEVEL
                self._full_path = {}
            else:
                # _LOG.debug('par ' + str(self.par_addr.path))
                self._full_path = dict(self.par_addr.get_full_path())
            self._full_path['@top'] = _NEXEL.CODE_TO_TOP_ENTITY_NAME[self.code]
            if self.obj_nex_id is not None:
                self._full_path['@idref'] = self.obj_nex_id
                other_id_key = _NEXEL.CODE_TO_OTHER_ID_KEY[self.code]
                if other_id_key is not None:
                    self._full_path[other_id_key] = self.obj_nex_id
            elif '@idref' in self._full_path:
                del self._full_path['@idref']
        return self._full_path

    @property
    def path(self):
        if self._path is None:
            if _USING_IDREF_ONLY_PATHS:
                if self.obj_nex_id is not None:
                    self._path = {'@idref': self.obj_nex_id}
                else:
                    if self.par_addr is None:
                        assert self.code == _NEXEL.TOP_LEVEL
                        self._path = {}
                    else:
                        # _LOG.debug('par ' + str(self.par_addr.path))
                        self._path = dict(self.par_addr.get_full_path())
                    self._path['@top'] = _NEXEL.CODE_TO_TOP_ENTITY_NAME[self.code]
                    if '@idref' in self._path:
                        del self._path['@idref']
            else:
                # _LOG.debug('c = ' + str(self.code))
                if self.par_addr is None:
                    if self.code not in [None, _NEXEL.TOP_LEVEL]:
                        _LOG.debug('code = ' + _NEXEL.CODE_TO_STR[self.code])
                        assert self.code == _NEXEL.TOP_LEVEL
                    self._path = {}
                else:
                    # _LOG.debug('par ' + str(self.par_addr.path))
                    self._path = dict(self.par_addr.path)
                self._path['@top'] = _NEXEL.CODE_TO_TOP_ENTITY_NAME[self.code]
                if self.obj_nex_id is not None:
                    self._path['@idref'] = self.obj_nex_id
                    other_id_key = _NEXEL.CODE_TO_OTHER_ID_KEY[self.code]
                    if other_id_key is not None:
                        assert not isinstance(self.obj_nex_id, dict)
                        self._path[other_id_key] = self.obj_nex_id
                elif '@idref' in self._path:
                    del self._path['@idref']
        return self._path


class _ValidationContext(object):
    """Holds references to the adaptor and logger
    """
    _et2schema_name = {
        _NEXEL.NEXML: '_NexmlEl_Schema',
        _NEXEL.OTUS: '_OtusEl_Schema',
        _NEXEL.OTU: '_OtuEl_Schema',
        _NEXEL.TREES: '_TreesEl_Schema',
        _NEXEL.TREE: '_TreeEl_Schema',
        _NEXEL.NODE: '_NodeEl_Schema',
        _NEXEL.INTERNAL_NODE: '_IntNEl_Schema',
        _NEXEL.LEAF_NODE: '_LeafEl_Schema',
        _NEXEL.EDGE: '_EdgeEl_Schema',
    }

    def __init__(self, adaptor, logger):
        self.adaptor = adaptor
        self.logger = logger
        self.anc_list = []
        self.curr_element_type = _NEXEL.TOP_LEVEL
        self.schema = None
        self._element_type_stack = []
        self._schema_stack = []

    def push_context(self, element_type, new_par):
        assert len(new_par) == 2
        assert not isinstance(new_par[1], dict)
        self.anc_list.append(new_par)
        self.push_context_no_anc(element_type)

    def pop_context(self):
        prev_anc = self.anc_list.pop(-1)
        et = self.pop_context_no_anc()
        return et, prev_anc

    def push_context_no_anc(self, element_type):
        self._element_type_stack.append(self.curr_element_type)
        self._schema_stack.append(self.schema)
        self.curr_element_type = element_type
        self.schema = getattr(self, _ValidationContext._et2schema_name[element_type])
        assert self.schema is not None

    def pop_context_no_anc(self):
        et = self.curr_element_type
        self.curr_element_type = self._element_type_stack.pop(-1)
        self.schema = self._schema_stack.pop(-1)
        return et

    def schema_name(self):
        names = ['_NexmlEl_Schema',
                 '_OtusEl_Schema',
                 '_OtuEl_Schema',
                 '_TreesEl_Schema',
                 '_TreeEl_Schema',
                 '_LeafEl_Schema',
                 '_IntNEl_Schema',
                 '_EdgeEl_Schema']
        for n in names:
            if self.schema is getattr(self, n, -1):
                return n
        return ''


class NexsonAnnotationAdder(object):
    def add_or_replace_annotation(self,  # pylint: disable=R0201
                                  obj,
                                  annotation,
                                  agent,
                                  add_agent_only=False):
        """Takes an `annotation` dictionary which is
        expected to have a string as the value of annotation['author']['name']
        This function will remove all annotations from obj that:
            1. have the same author/name, and
            2. have no messages that are flagged as messages to be preserved (values for 'preserve'
                that evaluate to true)
        """
        nex = get_nexml_el(obj)
        nvers = detect_nexson_version(obj)
        _LOG.debug('detected version as ' + nvers)
        agents_obj = find_val_literal_meta_first(nex, 'ot:agents', nvers)
        if not agents_obj:
            agents_obj = add_literal_meta(nex, 'ot:agents', {'agent': []}, nvers)
        agents_list = agents_obj.setdefault('agent', [])
        found_agent = False
        aid = agent['@id']
        for a in agents_list:
            if a.get('@id') == aid:
                found_agent = True
                break
        if not found_agent:
            agents_list.append(agent)
        if add_agent_only:
            delete_same_agent_annotation(obj, annotation)
        else:
            replace_same_agent_annotation(obj, annotation)


# noinspection PyUnusedLocal
class NexsonValidationAdaptor(NexsonAnnotationAdder):  # pylint: disable=R0921
    """An object created during NexSON validation.
    It holds onto the nexson object that it was instantiated for.
    When add_or_replace_annotation is called, it will annotate the
    nexson object, and when get_nexson_str is called it will
    serialize it.

    This class is useful merely because it allows the validation log
        and annotations to be relatively light weight, and yet easy
        to efficiently add back to the orignal NexSON object.

    Currently the only kwargs used is 'max_num_trees_per_study'
    """

    # noinspection PyTypeChecker
    def __init__(self, obj, logger, **kwargs):
        self._raw = obj
        self._nexml = None
        self._pyid_to_nexson_add = {}
        self._logger = logger
        self._repeated_id = False
        self._otuid2ottid_byogid = {}
        self._ottid2otuid_list_byogid = {}
        self._dupottid_by_ogid_tree_id = {}
        self._max_num_trees_per_study = kwargs.get('max_num_trees_per_study')
        uk = None
        for k in obj.keys():
            if k not in ['nexml', 'nex:nexml']:
                if uk is None:
                    uk = []
                uk.append(k)
        if uk:
            uk.sort()
            self._warn_event(_NEXEL.TOP_LEVEL,
                             obj=obj,
                             err_type=gen_UnrecognizedKeyWarning,
                             anc=_EMPTY_TUPLE,
                             obj_nex_id=None,
                             key_list=uk)
        self._nexml = None
        try:
            self._nexml = get_nexml_el(obj)
            assert isinstance(self._nexml, dict)
        except:
            self._error_event(_NEXEL.TOP_LEVEL,
                              obj=obj,
                              err_type=gen_MissingMandatoryKeyWarning,
                              anc=_EMPTY_TUPLE,
                              obj_nex_id=None,
                              key_list=['nexml', ])
            return  # EARLY EXIT!!
        self._nexson_id_to_obj = {}
        self._nexson_version = detect_nexson_version(obj)

        # attr used in validation only should be cleaned up
        # in the finally clause
        self._otu_group_by_id = {}
        self._otu_by_otug = {}

        try:
            # a little duck-punching
            vc = _ValidationContext(self, logger)
            try:
                add_schema_attributes(vc, self._nexson_version)
                assert self._nexson_version[:3] in ('0.0', '1.0', '1.2')
                self._validate_nexml_obj(self._nexml, vc, obj)
                if self._max_num_trees_per_study is not None:
                    nt = count_num_trees(self._raw)
                    if nt > self._max_num_trees_per_study:
                        m = '{f:d} trees found, but a limit of {m:d} trees per nexson is being enforced'
                        m = m.format(f=nt, m=self._max_num_trees_per_study)
                        self._error_event(_NEXEL.TOP_LEVEL,
                                          obj=obj,
                                          err_type=gen_MaxSizeExceededWarning,
                                          anc=_EMPTY_TUPLE,
                                          obj_nex_id=None,
                                          message=m)
                        return  # EARLY EXIT!!
            finally:
                vc.adaptor = None  # delete circular ref to help gc
                del vc
        finally:
            del self._otu_group_by_id
            del self._otu_by_otug

    def _fill_otu_ottid_maps(self, otus_group_id):
        if self._otuid2ottid_byogid.get(otus_group_id) is None:
            otuid2ottid = {}
            ottid2otuid_list = {}
            self._otuid2ottid_byogid[otus_group_id] = otuid2ottid
            self._ottid2otuid_list_byogid[otus_group_id] = ottid2otuid_list
            otu_dict = self._otu_by_otug[otus_group_id]
            for otuid, otu in otu_dict.items():
                ottid = find_val_literal_meta_first(otu, 'ot:ottId', self._nexson_version)
                otuid2ottid[otuid] = ottid
                ottid2otuid_list.setdefault(ottid, []).append(otuid)
            return otuid2ottid, ottid2otuid_list
        return self._otuid2ottid_byogid[otus_group_id], self._ottid2otuid_list_byogid[otus_group_id]

    def _detect_multilabelled_tree(self,
                                   otus_group_id,
                                   tree_id,
                                   otuid2leaf):
        # See if there are any otus that we need to flag as occurring in a tree
        # multiple_times
        #
        pair = self._fill_otu_ottid_maps(otus_group_id)
        ottid2otuid_list = pair[1]
        dup_dict = {}
        nd_list = None
        for ottid, otuid_list in ottid2otuid_list.items():
            if isinstance(otuid_list, list) and len(otuid_list) > 1:
                if nd_list is None:
                    nd_list = []
                for otuid in otuid_list:
                    nd_id = otuid2leaf.get(otuid)
                    if nd_id is not None:
                        nd_list.append((otuid, nd_id))
                if len(nd_list) > 1:
                    dup_dict[ottid] = nd_list
                    nd_list = None
                else:
                    del nd_list[:]
        bt = self._dupottid_by_ogid_tree_id.setdefault(otus_group_id, {})
        bt[tree_id] = dup_dict

    def _generate_ott_warnings(self, ogid2og_map, used_tree_id_list, nex_tuple, vc):
        for ogid, by_tree in self._dupottid_by_ogid_tree_id.items():
            ottid2otuid_list = self._ottid2otuid_list_byogid[ogid]
            dup_ottid_set = set()
            for ottid, otuid_list in ottid2otuid_list.items():
                if isinstance(otuid_list, list) and len(otuid_list):
                    dup_ottid_set.add(ottid)
            otuid2dup_set = set()
            for dottid in dup_ottid_set:
                for tree_id in used_tree_id_list:
                    dup_dict = by_tree.get(tree_id, {})
                    nl = dup_dict.get(dottid)
                    if nl and len(nl) > 1:
                        otu_ids = [i[0] for i in nl]  # (otu_id, no_id) pairs in nl
                        sotu = frozenset(otu_ids)
                        if sotu not in otuid2dup_set:
                            otuid2dup_set.add(sotu)
            if otuid2dup_set:
                vc.push_context(_NEXEL.OTUS, nex_tuple)
                try:
                    og = ogid2og_map[ogid]
                    for otu_set in otuid2dup_set:
                        self._warn_event(_NEXEL.OTUS,
                                         obj=og,
                                         err_type=gen_MultipleTipsToSameOttIdWarning,
                                         anc=vc.anc_list,
                                         obj_nex_id=ogid,
                                         otu_sets=list(otu_set))
                finally:
                    vc.pop_context()

    def _bf_meta_list_to_dict(self, m_list, par, par_vc):
        d = {}
        unparseable_m = None
        for m_el in m_list:
            try:
                n = m_el.get('@property')
            except:
                if unparseable_m is None:
                    unparseable_m = []
                unparseable_m.append(m_el)
            else:
                if n is None:
                    n = m_el.get('@rel')
                if n is None:
                    if unparseable_m is None:
                        unparseable_m = []
                    unparseable_m.append(m_el)
                else:
                    _add_value_to_dict_bf(d, n, m_el)
        if unparseable_m:
            for _ in unparseable_m:
                self._error_event(par_vc.curr_element_type,
                                  obj=par,
                                  err_type=gen_UnparseableMetaWarning,
                                  anc=par_vc.anc_list,
                                  obj_nex_id=None,
                                  obj_list=unparseable_m)
        return d

    _LIST_ADDR_INCR = 0

    def _event_address(self, element_type, obj, anc, obj_nex_id, anc_offset=0):
        if isinstance(obj_nex_id, list):
            obj_nex_id.sort()
            pyid = (NexsonValidationAdaptor._LIST_ADDR_INCR, None)
            NexsonValidationAdaptor._LIST_ADDR_INCR += 1
        else:
            pyid = id(obj)
        addr = self._pyid_to_nexson_add.get(pyid)
        if addr is None:
            if len(anc) > anc_offset:
                p_ind = -1 - anc_offset
                p, pnid = anc[p_ind]
                # _LOG.debug('addr is None branch... anc = ' + str(anc))
                pea = self._event_address(element_type=self._get_par_element_type(element_type),
                                          obj=p,
                                          anc=anc,
                                          obj_nex_id=pnid,
                                          anc_offset=1 + anc_offset)
                par_addr = pea[0]
            else:
                par_addr = None
            addr = LazyAddress(element_type, obj=obj, obj_nex_id=obj_nex_id, par_addr=par_addr)
            self._pyid_to_nexson_add[pyid] = addr
        return addr, pyid

    def _warn_event(self, element_type, obj, err_type, anc, obj_nex_id, *valist, **kwargs):
        c = factory2code[err_type]
        if not self._logger.is_logging_type(c):
            return
        address, pyid = self._event_address(element_type, obj, anc, obj_nex_id)
        err_type(address, pyid, self._logger, SeverityCodes.WARNING, *valist, **kwargs)

    def _error_event(self, element_type, obj, err_type, anc, obj_nex_id, *valist, **kwargs):
        c = factory2code[err_type]
        if not self._logger.is_logging_type(c):
            return
        # _LOG.debug('in _error_event = ' + str(obj_nex_id))
        address, pyid = self._event_address(element_type, obj, anc, obj_nex_id)
        # _LOG.debug('in _error_event address.obj_nex_id = ' + str(address.obj_nex_id))
        err_type(address, pyid, self._logger, SeverityCodes.ERROR, *valist, **kwargs)

    def _get_list_key(self, obj, key, vc, obj_nex_id=None):
        """Either:
            * Returns a list, or
            * Generates a MissingExpectedListWarning and returns None (if
                 the value is not a dict or list)
        """
        k = obj.get(key)
        if k is None:
            return None
        if isinstance(k, dict):
            k = [k]
        if not isinstance(k, list):
            self._error_event(vc.curr_element_type,
                              obj=obj,
                              err_type=gen_MissingExpectedListWarning,
                              anc=vc.anc_list,
                              obj_nex_id=obj_nex_id,
                              key_list=[key, ])
            return None
        return k

    def _get_par_element_type(self, c):  # pylint: disable=R0201
        pc = _NEXEL.CODE_TO_PAR_CODE.get(c)
        return pc

    def _check_meta_id(self, nid, meta_obj, k, container_obj, vc):  # pylint: disable=W0613
        robj = self._nexson_id_to_obj.setdefault(nid, meta_obj)
        if robj is meta_obj:
            return True
        self._error_event(vc.curr_element_type,
                          obj=meta_obj,
                          err_type=gen_RepeatedIDWarning,
                          anc=vc.anc_list,
                          obj_nex_id=nid,
                          key_list=[k])
        self._repeated_id = True
        return False

    def _register_nexson_id(self, nid, nobj, vc):
        robj = self._nexson_id_to_obj.setdefault(nid, nobj)
        if robj is nobj:
            return True
        self._error_event(vc.curr_element_type,
                          obj=nobj,
                          err_type=gen_RepeatedIDWarning,
                          anc=vc.anc_list,
                          obj_nex_id=nid,
                          key_list=[nid])
        self._repeated_id = True
        return False

    def _validate_obj_by_schema(self, obj, obj_nex_id, vc):
        """Creates:
            errors if `obj` does not contain keys in the schema.ALLOWED_KEY_SET,
            warnings if `obj` lacks keys listed in schema.EXPECETED_KEY_SET,
                      or if `obj` contains keys not listed in schema.ALLOWED_KEY_SET.
        """
        return self._validate_id_obj_list_by_schema([(obj_nex_id, obj)], vc, group_by_warning=False)

    def _validate_id_obj_list_by_schema(self, id_obj_list, vc, group_by_warning=False):
        element_type = vc.curr_element_type
        assert element_type is not None
        schema = vc.schema
        anc_list = vc.anc_list
        # _LOG.debug('using schema type = ' + vc.schema_name())
        using_hbf_meta = vc._using_hbf_meta  # pylint: disable=W0212
        _by_warn_type = {}
        for obj_nex_id, obj in id_obj_list:
            wrong_type = []
            unrec_meta_keys = []
            unrec_non_meta_keys = []

            if using_hbf_meta:
                for k, v in obj.items():
                    is_meta = k[0] == '^'
                    if k not in schema.ALLOWED_KEY_SET:
                        if is_meta:
                            unrec_meta_keys.append(k)
                        else:
                            unrec_non_meta_keys.append(k)
                    else:
                        correct_type, info = schema.K2VT[k](v, obj, k, vc)
                        if not correct_type:
                            wrong_type.append((k, v, info))
            else:
                for k, v in obj.items():
                    if k not in schema.ALLOWED_KEY_SET:
                        unrec_non_meta_keys.append(k)
                    else:
                        correct_type, info = schema.K2VT[k](v, obj, k, vc)
                        if not correct_type:
                            wrong_type.append((k, v, info))
                m = self._get_list_key(obj, 'meta', vc)
                if m:
                    # might want a flag of meta?
                    md = self._bf_meta_list_to_dict(m, obj, vc)
                    # _LOG.debug('md = ' + str(md))
                    mrmk = [i for i in schema.REQUIRED_META_KEY_SET if i not in md]
                    memk = [i for i in schema.EXPECTED_META_KEY_SET if i not in md]
                    if memk:
                        if group_by_warning:
                            memk.sort()
                            foks = frozenset(memk)
                            t = _by_warn_type.setdefault(foks, [[], []])
                            t[0].append(obj)
                            t[1].append(obj_nex_id)
                        else:
                            self._warn_event(element_type,
                                             obj=obj,
                                             err_type=gen_MissingOptionalKeyWarning,
                                             anc=anc_list,
                                             obj_nex_id=obj_nex_id,
                                             key_list=memk)
                    if mrmk:
                        self._error_event(element_type,
                                          obj=obj,
                                          err_type=gen_MissingMandatoryKeyWarning,
                                          anc=anc_list,
                                          obj_nex_id=obj_nex_id,
                                          key_list=mrmk)
                        msgf = 'missing mandatory meta key(s) according to {s} schema'
                        msg = msgf.format(s=vc.schema_name())
                        return error_return(msg)
                    for k, v in md.items():
                        if k not in schema.ALLOWED_META_KEY_SET:
                            unrec_meta_keys.append(k)
                        else:
                            # _LOG.debug('{k} --> "{v}"'.format(k=k, v=repr(v)))
                            correct_type, info = schema.K2VT[k](v, obj, k, vc)
                            if not correct_type:
                                v = extract_meta(v)
                                wrong_type.append((k, v, info))
            if wrong_type:
                self._error_event(element_type,
                                  obj=obj,
                                  err_type=gen_WrongValueTypeWarning,
                                  anc=anc_list,
                                  obj_nex_id=obj_nex_id,
                                  key_val_type_list=wrong_type)
                return error_return('wrong value type according to {s} schema'.format(s=vc.schema_name()))
            if unrec_non_meta_keys:
                self._warn_event(element_type,
                                 obj=obj,
                                 err_type=gen_UnrecognizedKeyWarning,
                                 anc=anc_list,
                                 obj_nex_id=obj_nex_id,
                                 key_list=unrec_non_meta_keys)
            if unrec_meta_keys:
                # might want a flag of meta?
                self._warn_event(element_type,
                                 obj=obj,
                                 err_type=gen_UnrecognizedKeyWarning,
                                 anc=anc_list,
                                 obj_nex_id=obj_nex_id,
                                 key_list=unrec_meta_keys)
            off_key = [k for k in schema.EXPECETED_KEY_SET if k not in obj]
            if off_key:
                if group_by_warning:
                    off_key.sort()
                    foks = frozenset(off_key)
                    t = _by_warn_type.setdefault(foks, [[], []])
                    t[0].append(obj)
                    t[1].append(obj_nex_id)
                else:
                    self._warn_event(element_type,
                                     obj=obj,
                                     err_type=gen_MissingOptionalKeyWarning,
                                     anc=anc_list,
                                     obj_nex_id=obj_nex_id,
                                     key_list=off_key)
            off_key = [k for k in schema.REQUIRED_KEY_SET if k not in obj]
            if off_key:
                self._error_event(element_type,
                                  obj=obj,
                                  err_type=gen_MissingMandatoryKeyWarning,
                                  anc=anc_list,
                                  obj_nex_id=obj_nex_id,
                                  key_list=off_key)
                return error_return('missing key(s) according to {s} schema'.format(s=vc.schema_name()))
        if _by_warn_type:
            for ks, obj_lists in _by_warn_type.items():
                mlist = list(ks)
                mlist.sort()
                id_arg = obj_lists[1]
                if len(id_arg) == 1:
                    id_arg = id_arg[0]
                obj_arg = obj_lists[0]
                if len(obj_arg) == 1:
                    obj_arg = obj_arg[0]
                self._warn_event(element_type,
                                 obj=obj_arg,
                                 err_type=gen_MissingOptionalKeyWarning,
                                 anc=anc_list,
                                 obj_nex_id=id_arg,
                                 key_list=mlist)
        return True

    def _validate_nexml_obj(self, nex_obj, vc, top_obj):
        vc.push_context(_NEXEL.NEXML, (top_obj, None))
        try:
            nid = nex_obj.get('@id')
            if nid is not None:
                if not self._register_nexson_id(nid, nex_obj, vc):
                    return False
            if not self._validate_obj_by_schema(nex_obj, nid, vc):
                return False
            return self._post_key_check_validate_nexml_obj(nex_obj, nid, vc)
        finally:
            vc.pop_context()

    def _validate_otus_group_list(self, otu_group_id_obj_list, vc):
        if not self._register_nexson_id_list(otu_group_id_obj_list, vc):
            return False
        for el in otu_group_id_obj_list:
            ogid, og = el
            self._otu_group_by_id[ogid] = og
            if not self._validate_obj_by_schema(og, ogid, vc):
                return False
            if not self._post_key_check_validate_otus_obj(ogid, og, vc):
                return False
        return True

    def _validate_trees_group_list(self, trees_group_id_obj_list, vc):
        if not self._register_nexson_id_list(trees_group_id_obj_list, vc):
            return False
        for el in trees_group_id_obj_list:
            tgid, tg = el
            if not self._validate_obj_by_schema(tg, tgid, vc):
                return False
            if not self._post_key_check_validate_tree_group(tgid, tg, vc):
                return False
        return True

    def _validate_tree(self, tree_id, tree_obj, vc, otus_group_id=None):
        if not self._register_nexson_id(tree_id, tree_obj, vc):
            return False
        if not self._validate_obj_by_schema(tree_obj, tree_id, vc):
            return False
        return self._post_key_check_validate_tree(tree_id,
                                                  tree_obj,
                                                  vc,
                                                  otus_group_id=otus_group_id)

    def _register_nexson_id_list(self, id_obj_list, vc):
        for obj_id, obj in id_obj_list:
            if not self._register_nexson_id(obj_id, obj, vc):
                return False
        return True

    def _validate_leaf_list(self, leaf_id_obj_list, vc):
        vc.push_context_no_anc(_NEXEL.LEAF_NODE)
        try:
            if not self._register_nexson_id_list(leaf_id_obj_list, vc):
                return False
            return self._validate_id_obj_list_by_schema(leaf_id_obj_list, vc)
        finally:
            vc.pop_context_no_anc()

    def _validate_internal_node_list(self, node_id_obj_list, vc):
        vc.push_context_no_anc(_NEXEL.INTERNAL_NODE)
        try:
            if not self._register_nexson_id_list(node_id_obj_list, vc):
                return False
            return self._validate_id_obj_list_by_schema(node_id_obj_list, vc)
        finally:
            vc.pop_context_no_anc()

    def _validate_node_list(self, node_id_obj_list, vc):
        vc.push_context_no_anc(_NEXEL.NODE)
        try:
            if not self._register_nexson_id_list(node_id_obj_list, vc):
                return False
            return self._validate_id_obj_list_by_schema(node_id_obj_list, vc)
        finally:
            vc.pop_context_no_anc()

    def _validate_edge_list(self, edge_id_obj_list, vc):
        vc.push_context_no_anc(_NEXEL.EDGE)
        try:
            if not self._register_nexson_id_list(edge_id_obj_list, vc):
                return False
            return self._validate_id_obj_list_by_schema(edge_id_obj_list, vc)
        finally:
            vc.pop_context_no_anc()

    def _validate_otu_list(self, otu_id_obj_list, vc):
        if not self._register_nexson_id_list(otu_id_obj_list, vc):
            return False
        # _LOG.debug(str(otu_id_obj_list))
        if not self._validate_id_obj_list_by_schema(otu_id_obj_list, vc, group_by_warning=True):
            return False
        return self._post_key_check_validate_otu_id_obj_list(otu_id_obj_list, vc)

    def _post_key_check_validate_otu_id_obj_list(self, otu_id_obj_list, vc):  # pylint: disable=R0201, W0613
        return True

    def _post_key_check_validate_tree(self,
                                      tree_nex_id,
                                      tree_obj,
                                      vc,
                                      otus_group_id=None):
        raise NotImplementedError('base NexsonValidationAdaptor hook')

    def _post_key_check_validate_nexml_obj(self, nex_obj, obj_nex_id, vc):
        raise NotImplementedError('base NexsonValidationAdaptor hook')

    def _post_key_check_validate_otus_obj(self, og_nex_id, otus_group, vc):
        raise NotImplementedError('base NexsonValidationAdaptor hook')

    def _post_key_check_validate_tree_group(self, tg_nex_id, trees_group, vc):
        raise NotImplementedError('base NexsonValidationAdaptor hook')

    def get_nexson_str(self):
        return json.dumps(self._raw, sort_keys=True, indent=0)

    def _validate_otu_key_if_present(self, nd_id_nd_iter, otu_by_id, vc):
        erred = False
        for nd_id, nd in nd_id_nd_iter:
            o = nd.get('@otu')
            if o is not None:
                if o not in otu_by_id:
                    erred = True
                    self._error_event(_NEXEL.NODE,
                                      obj=nd,
                                      err_type=gen_ReferencedIDNotFoundWarning,
                                      anc=vc.anc_list,
                                      obj_nex_id=nd_id,
                                      key_list=[o])
        if erred:
            error_return('Unknown "@otu" id')
        return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
nexson testing suite.
"""
##############################################################################
#  Adapted from: DendroPy Phylogenetic Computing Library.
#
#  Copyright 2010 Jeet Sukumaran and Mark T. Holder.
#  All rights reserved.
#
#  Sukumaran, J. and M. T. Holder. 2010. DendroPy: a Python library
#     for phylogenetic computing. Bioinformatics 26: 1569-1571.
#
##############################################################################

import unittest
import re
import os

from .support.pathmap import get_test_path_mapper

def get_test_file_names():
    """Get list of test file names."""
    path = os.path.dirname(__file__)
    files = os.listdir(path)
    t = []
    pat = re.compile(r'^test.*\.py$')
    for f in files:
        if pat.match(f):
            rp = 'nexson.test.' + f[:-3]  # [:-3] to strip ".py"
            t.append(rp)
    return t


def get_test_suite(test_file_names=None):
    """
    Creates a unittest.TestSuite from all of the modules in
    `nexson.test`. Right now, assumes (a) no subdirectories (though
    this can easily be accommodated) and (b) every test to be run is
    sitting in a module with a file name of 'test*.py', and, conversely,
    every file with a name of 'test*.py' has test(s) to be run.
    """
    if test_file_names is None:
        test_file_names = get_test_file_names()
    tests = unittest.defaultTestLoader.loadTestsFromNames(test_file_names)
    return unittest.TestSuite(tests)


def run():
    """Runs all of the unittests"""
    runner = unittest.TextTestRunner()
    runner.run(get_test_suite())


if __name__ == "__main__":
    run()



#!/usr/bin/env python
from ..syntax.helper import (extract_meta,
                             _is_badgerfish_version,
                             _is_by_id_hbf,
                             _is_direct_hbf)
from peyutil import is_str_type
import logging

_LOG = logging.getLogger(__name__)


# pylint: disable=W0613,W0212
def add_schema_attributes(container, nexson_version):
    """Adds several attributes to `container`:
    _using_hbf_meta  - boolean. True for HoneyBadgerFish v1-style meta elements ('^prop': value
                rather than 'meta': {'$':value})
    and the following _SchemaFragment instances:
    _NexmlEl_Schema
    """
    if _is_by_id_hbf(nexson_version):
        _add_by_id_nexson_schema_attributes(container)
    elif _is_badgerfish_version(nexson_version):
        _add_badgerfish_nexson_schema_attributes(container)
    elif _is_direct_hbf(nexson_version):
        _add_direct_nexson_schema_attributes(container)
    else:
        raise NotImplementedError('unrecognized nexson variant {}'.format(nexson_version))


class _SchemaFragment(object):
    def __init__(self,
                 required,
                 expected,
                 allowed,
                 required_meta,
                 expected_meta,
                 type_checked_meta,
                 using_hbf_meta):
        """
        `required` dict of key to _VT type code. Error is not present.
        `expected` dict of key to _VT type code. Warnings is not present.
        `allowed` dict of None ignored elements #@TODO could be a set.
        `required_meta` dict of meta key to _VT type code. Error is not present.
        `expected_meta` dict of meta key to _VT type code. Warnings is not present.
        `type_checked_meta` dict of meta key to _VT type code. Error if wrong type, no warning if absent.
        `using_hbf_meta` True for >=1.0.0, False for badgerfish.
        """
        self.K2VT = {}
        for k, v in required.items():
            self.K2VT[k] = _VT.m2RawCheck[v]
        for k, v in expected.items():
            self.K2VT[k] = _VT.m2RawCheck[v]
        for k, v in allowed.items():
            self.K2VT[k] = _VT.m2RawCheck[v]
        required_meta_keys = tuple(required_meta.keys())
        expected_meta_keys = tuple(expected_meta.keys())
        required_keys = tuple(required.keys())
        expected_keys = tuple(expected.keys())
        allowed_keys = tuple(allowed.keys())
        if using_hbf_meta:
            for k, v in required_meta.items():
                self.K2VT['^' + k] = _VT._2HBFCheck[v]
            for k, v in expected_meta.items():
                self.K2VT['^' + k] = _VT._2HBFCheck[v]
            for k, v in type_checked_meta.items():
                self.K2VT['^' + k] = _VT._2HBFCheck[v]
            self.REQUIRED_META_KEY_SET = frozenset(['^' + i for i in required_meta_keys])
            self.EXPECTED_META_KEY_SET = frozenset(['^' + i for i in expected_meta_keys])
            self.REQUIRED_KEY_SET = frozenset(required_keys + tuple(self.REQUIRED_META_KEY_SET))
            self.EXPECETED_KEY_SET = frozenset(expected_keys + tuple(self.EXPECTED_META_KEY_SET))
            self.ALLOWED_KEY_SET = frozenset(tuple(self.REQUIRED_KEY_SET)
                                             + tuple(self.EXPECETED_KEY_SET)
                                             + allowed_keys
                                             + tuple(['^' + i for i in type_checked_meta.keys()]))
        else:
            for k, v in required_meta.items():
                self.K2VT[k] = _VT._2BFCheck[v]
            for k, v in expected_meta.items():
                self.K2VT[k] = _VT._2BFCheck[v]
            for k, v in type_checked_meta.items():
                self.K2VT[k] = _VT._2BFCheck[v]
            self.K2VT['meta'] = _VT.m2RawCheck[_VT.LIST_OR_DICT]
            self.REQUIRED_META_KEY_SET = frozenset(required_meta_keys)
            self.EXPECTED_META_KEY_SET = frozenset(expected_meta_keys)
            targ = required_meta_keys + expected_meta_keys + tuple(type_checked_meta.keys())
            self.ALLOWED_META_KEY_SET = frozenset(targ)
            self.REQUIRED_KEY_SET = frozenset(required_keys)
            self.EXPECETED_KEY_SET = frozenset(expected_keys)
            self.ALLOWED_KEY_SET = frozenset(('meta',)
                                             + tuple(self.REQUIRED_KEY_SET)
                                             + tuple(self.EXPECETED_KEY_SET)
                                             + allowed_keys)
        # for order-dependent
        x = list(self.REQUIRED_META_KEY_SET)
        x.sort()
        self.REQUIRED_META_KEY_TUPLE = tuple(x)
        x = list(self.EXPECTED_META_KEY_SET)
        x.sort()
        self.EXPECTED_META_KEY_TUPLE = tuple(x)
        x = list(self.ALLOWED_KEY_SET)
        x.sort()
        self.ALLOWED_KEY_TUPLE = tuple(x)
        x = list(self.REQUIRED_META_KEY_SET)
        x.sort()
        self.REQUIRED_META_TUPLE = tuple(x)
        x = list(self.EXPECTED_META_KEY_SET)
        x.sort()
        self.EXPECTED_META_KEY_TUPLE = tuple(x)
        self.USING_HBF_META = using_hbf_meta


__TRUE_VAL = (True, None)
__FALSE_STR = (False, 'str')
__FALSE_HREF = (False, 'href')
__FALSE_LIST = (False, 'array')
__FALSE_DICT = (False, 'object')
__FALSE_STR_REPEATABLE_EL = (False, 'string or string array')
__FALSE_STR_LIST = (False, 'string array')
__FALSE_INT = (False, 'integer')
__FALSE_FLOAT = (False, 'float')
__FALSE_DICT_LIST = (False, 'array or object')
__FALSE_BOOL = (False, 'boolean')


def _check_id(x, obj, k, vc):
    try:
        nid = x.get('@id')
    except:
        return
    if nid is not None:
        vc.adaptor._check_meta_id(nid, x, k, obj, vc)


# noinspection PyUnusedLocal
def check_raw_bool(x, obj, k, vc):
    return __TRUE_VAL if (x is False or x is True) else __FALSE_BOOL


def check_obj_meta_bool(x, obj, k, vc):
    mo = extract_meta(x)
    _check_id(x, obj, k, vc)
    return check_raw_bool(mo, obj, k, vc)


def check_hbf_meta_bool(x, obj, k, vc):
    if isinstance(x, dict):
        return check_obj_meta_bool(x, obj, k, vc)
    return __TRUE_VAL if (x is False or x is True) else __FALSE_BOOL


# noinspection PyUnusedLocal
def check_raw_dict(x, obj, k, vc):
    return __TRUE_VAL if (isinstance(x, dict)) else __FALSE_DICT


def check_obj_meta_dict(x, obj, k, vc):
    mo = extract_meta(x)
    _check_id(x, obj, k, vc)
    return check_raw_dict(mo, obj, k, vc)


# noinspection PyUnusedLocal
def check_hbf_meta_dict(x, obj, k, vc):
    if isinstance(x, dict):
        return __TRUE_VAL


def check_href(x, obj, k, vc):
    try:
        _check_id(x, obj, k, vc)
        h = x.get('@href')
        if is_str_type(h):
            return __TRUE_VAL
    except:
        pass
    return __FALSE_HREF


# noinspection PyUnusedLocal
def check_raw_int(x, obj, k, vc):
    return __TRUE_VAL if (isinstance(x, int)) else __FALSE_INT


def check_obj_meta_int(x, obj, k, vc):
    mo = extract_meta(x)
    _check_id(x, obj, k, vc)
    return check_raw_int(mo, obj, k, vc)


def check_hbf_meta_int(x, obj, k, vc):
    if isinstance(x, dict):
        return check_obj_meta_int(x, obj, k, vc)
    return check_raw_int(x, obj, k, vc)


# noinspection PyUnusedLocal
def check_raw_float(x, obj, k, vc):
    return __TRUE_VAL if (isinstance(x, float) or isinstance(x, int)) else __FALSE_FLOAT


def check_obj_meta_float(x, obj, k, vc):
    mo = extract_meta(x)
    _check_id(x, obj, k, vc)
    return check_raw_float(mo, obj, k, vc)


def check_hbf_meta_float(x, obj, k, vc):
    if isinstance(x, dict):
        return check_obj_meta_float(x, obj, k, vc)
    return check_raw_float(x, obj, k, vc)


# noinspection PyUnusedLocal
def check_raw_list(x, obj, k, vc):
    return __TRUE_VAL if (isinstance(x, list)) else __FALSE_LIST


def check_obj_meta_list(x, obj, k, vc):
    mo = extract_meta(x)
    _check_id(x, obj, k, vc)
    return check_raw_list(mo, obj, k, vc)


def check_hbf_meta_list(x, obj, k, vc):
    if isinstance(x, dict):
        return check_obj_meta_list(x, obj, k, vc)
    return check_raw_list(x, obj, k, vc)


# noinspection PyUnusedLocal
def check_list_or_dict(x, obj, k, vc):
    return __TRUE_VAL if (isinstance(x, list) or isinstance(x, dict)) else __FALSE_DICT_LIST

# noinspection PyUnusedLocal
def check_raw_str(x, obj, k, vc):
    return __TRUE_VAL if is_str_type(x) else __FALSE_STR


def check_obj_meta_str(x, obj, k, vc):
    mo = extract_meta(x)
    _check_id(x, obj, k, vc)
    return check_raw_str(mo, obj, k, vc)


def check_hbf_meta_str(x, obj, k, vc):
    if isinstance(x, dict):
        return check_obj_meta_str(x, obj, k, vc)
    return check_raw_str(x, obj, k, vc)

# noinspection PyUnusedLocal
def check_raw_str_list(x, obj, k, vc):
    if not isinstance(x, list):
        return __FALSE_STR_LIST
    for i in x:
        if not is_str_type(i):
            return __FALSE_STR_LIST
    return __TRUE_VAL


def check_obj_meta_str_list(x, obj, k, vc):
    mo = extract_meta(x)
    _check_id(x, obj, k, vc)
    return check_raw_str_list(mo, obj, k, vc)


def check_hbf_meta_str_list(x, obj, k, vc):
    if isinstance(x, dict):
        return check_obj_meta_str_list(x, obj, k, vc)
    return check_raw_str_list(x, obj, k, vc)


# noinspection PyUnusedLocal
def check_raw_str_repeatable(x, obj, k, vc):
    if is_str_type(x):
        return __TRUE_VAL
    if not isinstance(x, list):
        return __FALSE_STR_REPEATABLE_EL
    for i in x:
        if not is_str_type(i):
            return __FALSE_STR_REPEATABLE_EL
    return __TRUE_VAL


def check_obj_meta_str_repeatable(x, obj, k, vc):
    if not isinstance(x, list):
        x = [x]
    for el in x:
        r = check_obj_meta_str(el, obj, k, vc)
        if r is not __TRUE_VAL:
            return __FALSE_STR_REPEATABLE_EL
    return __TRUE_VAL


def check_hbf_meta_str_repeatable(x, obj, k, vc):
    if not isinstance(x, list):
        x = [x]
    for el in x:
        r = check_hbf_meta_str(el, obj, k, vc)
        if r is not __TRUE_VAL:
            return __FALSE_STR_REPEATABLE_EL
    return __TRUE_VAL


# noinspection PyShadowingBuiltins
class _VT(object):
    """Value type enum"""
    BOOL = 0
    DICT = 1
    HREF = 2
    INT = 3
    LIST = 4
    LIST_OR_DICT = 5
    STR = 6
    STR_LIST = 7
    STR_REPEATABLE_EL = 8
    FLOAT = 9

    m2RawCheck = {
        BOOL: check_raw_bool,
        DICT: check_raw_dict,
        HREF: check_href,
        INT: check_raw_int,
        FLOAT: check_raw_float,
        LIST: check_raw_list,
        LIST_OR_DICT: check_list_or_dict,
        STR: check_raw_str,
        STR_LIST: check_raw_str_list,
        STR_REPEATABLE_EL: check_raw_str_repeatable,
    }
    _2HBFCheck = {
        BOOL: check_hbf_meta_bool,
        DICT: check_hbf_meta_dict,
        HREF: check_href,
        INT: check_raw_int,
        FLOAT: check_raw_float,
        LIST: check_hbf_meta_list,
        LIST_OR_DICT: check_list_or_dict,
        STR: check_hbf_meta_str,
        STR_LIST: check_hbf_meta_str_list,
        STR_REPEATABLE_EL: check_hbf_meta_str_repeatable,
    }

    _2BFCheck = {
        BOOL: check_obj_meta_bool,
        DICT: check_obj_meta_dict,
        HREF: check_href,
        INT: check_obj_meta_int,
        FLOAT: check_obj_meta_float,
        LIST: check_hbf_meta_list,
        LIST_OR_DICT: check_list_or_dict,
        STR: check_obj_meta_str,
        STR_LIST: check_obj_meta_str_list,
        STR_REPEATABLE_EL: check_obj_meta_str_repeatable,
    }


_SchemaFragment._VT = _VT

# In the "schema definitions" below, the names are concatenations:
# first:
#   _ to make the vars private
# second, an element to indicate when the element is expected:
#   Req = "required. error if not present"
#   Exp = "expected, warn if not present"
#   All = "allowed (no warning if absent or present"
#   Typ = "type-checked" like allowed (but only used for Meta)
# third:
#   _ = non-meta
#   M = meta
# fourth:
#   Parent element + El_ (NexmlEl_, OtusEl_...)
# last is mneumonic for the nexson variant:
#   All = all variants
#   Dir = 1.0.0 ("direct" hbf translation of nexml)
#   ByI = 1.2   (the "by ID" hbf)
#
_EMPTY_DICT = {}
####################################################################
# nexml element schema
_Req_NexmlEl_All = {'@id': _VT.STR,
                    }
_Exp_NexmlEl_Dir = {'otus': _VT.LIST_OR_DICT,
                    'trees': _VT.LIST_OR_DICT,
                    }
_Exp_NexmlEl_ByI = {'otusById': _VT.DICT,
                    'treesById': _VT.DICT,
                    '^ot:otusElementOrder': _VT.STR_LIST,
                    '^ot:treesElementOrder': _VT.STR_LIST,
                    }
_All_NexmlEl_All = {'@about': _VT.STR,
                    '@generator': _VT.STR,
                    '@nexmljson': _VT.STR,
                    '@version': _VT.STR,
                    '@xmlns': _VT.DICT,
                    '@nexml2json': _VT.STR,
                    }
_ExpMNexmlEl_All = {'ot:dataDeposit': _VT.HREF,
                    'ot:studyPublication': _VT.HREF,
                    'ot:studyPublicationReference': _VT.STR,
                    'ot:studyYear': _VT.INT,
                    }
_TypMNexmlEl_All = {'ot:annotationEvents': _VT.DICT,
                    'ot:agents': _VT.DICT,
                    'ot:curatorName': _VT.STR_REPEATABLE_EL,
                    'ot:focalClade': _VT.INT,
                    'ot:focalCladeOTTTaxonName': _VT.STR,
                    'ot:messages': _VT.DICT,
                    'ot:notIntendedForSynthesis': _VT.BOOL,
                    'ot:notUsingRootedTrees': _VT.BOOL,
                    'ot:studyId': _VT.STR,
                    'ot:tag': _VT.STR_REPEATABLE_EL,
                    'ot:taxonLinkPrefixes': _VT.DICT,
                    'ot:candidateTreeForSynthesis': _VT.STR_REPEATABLE_EL,
                    'xhtml:license': _VT.HREF,
                    }

_v1_2_nexml = _SchemaFragment(required=_Req_NexmlEl_All,
                              expected=_Exp_NexmlEl_ByI,
                              allowed=_All_NexmlEl_All,
                              required_meta=_EMPTY_DICT,
                              expected_meta=_ExpMNexmlEl_All,
                              type_checked_meta=_TypMNexmlEl_All,
                              using_hbf_meta=True)

_v1_0_nexml = _SchemaFragment(required=_Req_NexmlEl_All,
                              expected=_Exp_NexmlEl_Dir,
                              allowed=_All_NexmlEl_All,
                              required_meta=_EMPTY_DICT,
                              expected_meta=_ExpMNexmlEl_All,
                              type_checked_meta=_TypMNexmlEl_All,
                              using_hbf_meta=True)

_v0_0_nexml = _SchemaFragment(required=_Req_NexmlEl_All,
                              expected=_Exp_NexmlEl_Dir,
                              allowed=_All_NexmlEl_All,
                              required_meta=_EMPTY_DICT,
                              expected_meta=_ExpMNexmlEl_All,
                              type_checked_meta=_TypMNexmlEl_All,
                              using_hbf_meta=False)
#####################################################################

####################################################################
# otus element schema
_Req_OtusEl_ByI = {'otuById': _VT.DICT,
                   }
_Req_OtusEl_Dir = {'@id': _VT.STR,
                   'otu': _VT.LIST,
                   }
_All_OtusEl_Dir = {'@about': _VT.STR,
                   }

_v1_2_Otus = _SchemaFragment(required=_Req_OtusEl_ByI,
                             expected=_EMPTY_DICT,
                             allowed=_EMPTY_DICT,
                             required_meta=_EMPTY_DICT,
                             expected_meta=_EMPTY_DICT,
                             type_checked_meta=_EMPTY_DICT,
                             using_hbf_meta=True)

_v1_0_Otus = _SchemaFragment(required=_Req_OtusEl_Dir,
                             expected=_EMPTY_DICT,
                             allowed=_All_OtusEl_Dir,
                             required_meta=_EMPTY_DICT,
                             expected_meta=_EMPTY_DICT,
                             type_checked_meta=_EMPTY_DICT,
                             using_hbf_meta=True)

_v0_0_Otus = _SchemaFragment(required=_Req_OtusEl_Dir,
                             expected=_EMPTY_DICT,
                             allowed=_All_OtusEl_Dir,
                             required_meta=_EMPTY_DICT,
                             expected_meta=_EMPTY_DICT,
                             type_checked_meta=_EMPTY_DICT,
                             using_hbf_meta=False)

####################################################################
# otu element schema
_Req_OtuEl_ByI = _EMPTY_DICT
_Req_OtuEl_Dir = {'@id': _VT.STR,
                  }
_All_OtuEl_ByI = {'@label': _VT.STR,
                  }
_All_OtuEl_Dir = {'@about': _VT.STR,
                  '@label': _VT.STR,
                  }
_ReqMOtuEl_All = {'ot:originalLabel': _VT.STR,
                  }
_ExpMOtuEl_All = {'ot:ottId': _VT.INT,
                  }
_TypMOtuEl_All = {'ot:treebaseOTUId': _VT.STR,
                  'ot:ottTaxonName': _VT.STR,
                  'ot:altLabel': _VT.STR,
                  'ot:taxonLink': _VT.DICT,
                  "skos:altLabel": _VT.LIST_OR_DICT,
                  }
_v1_2_Otu = _SchemaFragment(required=_Req_OtuEl_ByI,
                            expected=_EMPTY_DICT,
                            allowed=_All_OtuEl_ByI,
                            required_meta=_ReqMOtuEl_All,
                            expected_meta=_ExpMOtuEl_All,
                            type_checked_meta=_TypMOtuEl_All,
                            using_hbf_meta=True)
_v1_0_Otu = _SchemaFragment(required=_Req_OtuEl_Dir,
                            expected=_EMPTY_DICT,
                            allowed=_All_OtuEl_Dir,
                            required_meta=_ReqMOtuEl_All,
                            expected_meta=_ExpMOtuEl_All,
                            type_checked_meta=_TypMOtuEl_All,
                            using_hbf_meta=True)

_v0_0_Otu = _SchemaFragment(required=_Req_OtuEl_Dir,
                            expected=_EMPTY_DICT,
                            allowed=_All_OtuEl_Dir,
                            required_meta=_ReqMOtuEl_All,
                            expected_meta=_ExpMOtuEl_All,
                            type_checked_meta=_TypMOtuEl_All,
                            using_hbf_meta=False)
#####################################################################

####################################################################
# trees element schema
_Req_TreesEl_ByI = {'@otus': _VT.STR,
                    'treeById': _VT.DICT,
                    '^ot:treeElementOrder': _VT.STR_LIST,
                    }
_Req_TreesEl_Dir = {'@id': _VT.STR,
                    '@otus': _VT.STR,
                    'tree': _VT.LIST_OR_DICT,
                    }
_All_TreesEl_Dir = {'@about': _VT.STR,
                    }

_v1_2_Trees = _SchemaFragment(required=_Req_TreesEl_ByI,
                              expected=_EMPTY_DICT,
                              allowed=_EMPTY_DICT,
                              required_meta=_EMPTY_DICT,
                              expected_meta=_EMPTY_DICT,
                              type_checked_meta=_EMPTY_DICT,
                              using_hbf_meta=True)

_v1_0_Trees = _SchemaFragment(required=_Req_TreesEl_Dir,
                              expected=_EMPTY_DICT,
                              allowed=_All_TreesEl_Dir,
                              required_meta=_EMPTY_DICT,
                              expected_meta=_EMPTY_DICT,
                              type_checked_meta=_EMPTY_DICT,
                              using_hbf_meta=True)

_v0_0_Trees = _SchemaFragment(required=_Req_TreesEl_Dir,
                              expected=_EMPTY_DICT,
                              allowed=_All_TreesEl_Dir,
                              required_meta=_EMPTY_DICT,
                              expected_meta=_EMPTY_DICT,
                              type_checked_meta=_EMPTY_DICT,
                              using_hbf_meta=False)
#####################################################################

####################################################################
# tree element schema
_Req_TreeEl_ByI = {'edgeBySourceId': _VT.DICT,
                   'nodeById': _VT.DICT,
                   }
_Req_TreeEl_Dir = {'edge': _VT.LIST,
                   'node': _VT.LIST,
                   '@id': _VT.STR,
                   }
_Exp_TreeEl_All = {'@xsi:type': _VT.STR,  # could be a choice...
                   }
_All_TreeEl_ByI = {'@label': _VT.STR,
                   }
_All_TreeEl_Dir = {'@about': _VT.STR,
                   '@label': _VT.STR,
                   }
_ReqMTreeEl_ByI = {'ot:rootNodeId': _VT.STR,
                   }
_ReqMTreeEl_Dir = _EMPTY_DICT
_ExpMTreeEl_All = {'ot:branchLengthDescription': _VT.STR,
                   'ot:branchLengthTimeUnit': _VT.STR,
                   'ot:branchLengthMode': _VT.STR,
                   'ot:curatedType': _VT.STR,
                   'ot:inGroupClade': _VT.STR,
                   'ot:outGroupEdge': _VT.STR,
                   }
_TypMTreeEl_All = {'ot:specifiedRoot': _VT.STR,
                   'ot:tag': _VT.STR_REPEATABLE_EL,
                   'ot:unrootedTree': _VT.BOOL,
                   }
_v1_2_Tree = _SchemaFragment(required=_Req_TreeEl_ByI,
                             expected=_Exp_TreeEl_All,
                             allowed=_All_TreeEl_ByI,
                             required_meta=_ReqMTreeEl_ByI,
                             expected_meta=_ExpMTreeEl_All,
                             type_checked_meta=_TypMTreeEl_All,
                             using_hbf_meta=True)
_v1_0_Tree = _SchemaFragment(required=_Req_TreeEl_Dir,
                             expected=_Exp_TreeEl_All,
                             allowed=_All_TreesEl_Dir,
                             required_meta=_ReqMTreeEl_Dir,
                             expected_meta=_ExpMTreeEl_All,
                             type_checked_meta=_TypMTreeEl_All,
                             using_hbf_meta=True)

_v0_0_Tree = _SchemaFragment(required=_Req_TreeEl_Dir,
                             expected=_Exp_TreeEl_All,
                             allowed=_All_TreesEl_Dir,
                             required_meta=_ReqMTreeEl_Dir,
                             expected_meta=_ExpMTreeEl_All,
                             type_checked_meta=_TypMTreeEl_All,
                             using_hbf_meta=False)
#####################################################################

####################################################################
# leaf node element schema
_Req_LeafEl_ByI = {'@otu': _VT.STR,
                   }
_Req_LeafEl_Dir = {'@id': _VT.STR,
                   '@otu': _VT.STR,  # could be a choice...
                   }
_All_LeafEl_Dir = {'@about': _VT.STR,
                   }
_ReqMLeafEl_Dir = {'ot:isLeaf': _VT.BOOL,
                   }
_v1_2_Leaf = _SchemaFragment(required=_Req_LeafEl_ByI,
                             expected=_EMPTY_DICT,
                             allowed=_EMPTY_DICT,
                             required_meta=_EMPTY_DICT,
                             expected_meta=_EMPTY_DICT,
                             type_checked_meta=_EMPTY_DICT,
                             using_hbf_meta=True)
_v1_0_Leaf = _SchemaFragment(required=_Req_LeafEl_Dir,
                             expected=_EMPTY_DICT,
                             allowed=_All_LeafEl_Dir,
                             required_meta=_ReqMLeafEl_Dir,
                             expected_meta=_EMPTY_DICT,
                             type_checked_meta=_EMPTY_DICT,
                             using_hbf_meta=True)
_v0_0_Leaf = _SchemaFragment(required=_Req_LeafEl_Dir,
                             expected=_EMPTY_DICT,
                             allowed=_All_LeafEl_Dir,
                             required_meta=_ReqMLeafEl_Dir,
                             expected_meta=_EMPTY_DICT,
                             type_checked_meta=_EMPTY_DICT,
                             using_hbf_meta=False)
#####################################################################

####################################################################
# internal node element schema
_Req_IntNEl_ByI = _EMPTY_DICT
_Req_IntNEl_Dir = {'@id': _VT.STR,
                   }
_All_IntNEl_ByI = {'@root': _VT.BOOL,
                   '@otu': _VT.STR
                   }
_All_IntNEl_Dir = {'@about': _VT.STR,
                   '@root': _VT.BOOL,
                   '@otu': _VT.STR
                   }

_TypMIntNEl_Dir = {'ot:isLeaf': _VT.BOOL,
                   }
_v1_2_IntN = _SchemaFragment(required=_Req_IntNEl_ByI,
                             expected=_EMPTY_DICT,
                             allowed=_All_IntNEl_ByI,
                             required_meta=_EMPTY_DICT,
                             expected_meta=_EMPTY_DICT,
                             type_checked_meta=_EMPTY_DICT,
                             using_hbf_meta=True)
_v1_0_IntN = _SchemaFragment(required=_Req_IntNEl_Dir,
                             expected=_EMPTY_DICT,
                             allowed=_All_IntNEl_Dir,
                             required_meta=_EMPTY_DICT,
                             expected_meta=_EMPTY_DICT,
                             type_checked_meta=_TypMIntNEl_Dir,
                             using_hbf_meta=True)
_v0_0_IntN = _SchemaFragment(required=_Req_IntNEl_Dir,
                             expected=_EMPTY_DICT,
                             allowed=_All_IntNEl_Dir,
                             required_meta=_EMPTY_DICT,
                             expected_meta=_EMPTY_DICT,
                             type_checked_meta=_TypMIntNEl_Dir,
                             using_hbf_meta=False)
#####################################################################

####################################################################
# node element schema
_Req_NodeEl_ByI = _EMPTY_DICT
_Req_NodeEl_Dir = {'@id': _VT.STR,
                   }
_All_NodeEl_ByI = {'@otu': _VT.STR,
                   '@root': _VT.BOOL,
                   }
_All_NodeEl_Dir = {'@about': _VT.STR,
                   '@otu': _VT.STR,
                   '@root': _VT.BOOL,
                   }
_TypMNodeEl_Dir = {'ot:isLeaf': _VT.BOOL,
                   }
_v1_2_Node = _SchemaFragment(required=_Req_NodeEl_ByI,
                             expected=_EMPTY_DICT,
                             allowed=_All_NodeEl_ByI,
                             required_meta=_EMPTY_DICT,
                             expected_meta=_EMPTY_DICT,
                             type_checked_meta=_EMPTY_DICT,
                             using_hbf_meta=True)
_v1_0_Node = _SchemaFragment(required=_Req_NodeEl_Dir,
                             expected=_EMPTY_DICT,
                             allowed=_All_NodeEl_Dir,
                             required_meta=_EMPTY_DICT,
                             expected_meta=_EMPTY_DICT,
                             type_checked_meta=_TypMNodeEl_Dir,
                             using_hbf_meta=True)
_v0_0_Node = _SchemaFragment(required=_Req_NodeEl_Dir,
                             expected=_EMPTY_DICT,
                             allowed=_All_NodeEl_Dir,
                             required_meta=_EMPTY_DICT,
                             expected_meta=_EMPTY_DICT,
                             type_checked_meta=_TypMNodeEl_Dir,
                             using_hbf_meta=False)
#####################################################################

####################################################################
# edge element schema
_Req_EdgeEl_ByI = {'@target': _VT.STR,
                   '@source': _VT.STR,
                   }
_Req_EdgeEl_Dir = {'@target': _VT.STR,
                   '@id': _VT.STR,
                   '@source': _VT.STR,
                   }
_All_EdgeEl_ByI = {'@length': _VT.FLOAT,
                   }
_All_EdgeEl_Dir = {'@about': _VT.STR,
                   '@length': _VT.FLOAT,
                   }
_v1_2_Edge = _SchemaFragment(required=_Req_EdgeEl_ByI,
                             expected=_EMPTY_DICT,
                             allowed=_All_EdgeEl_ByI,
                             required_meta=_EMPTY_DICT,
                             expected_meta=_EMPTY_DICT,
                             type_checked_meta=_EMPTY_DICT,
                             using_hbf_meta=True)
_v1_0_Edge = _SchemaFragment(required=_Req_EdgeEl_Dir,
                             expected=_EMPTY_DICT,
                             allowed=_All_EdgeEl_Dir,
                             required_meta=_EMPTY_DICT,
                             expected_meta=_EMPTY_DICT,
                             type_checked_meta=_EMPTY_DICT,
                             using_hbf_meta=True)
_v0_0_Edge = _SchemaFragment(required=_Req_EdgeEl_Dir,
                             expected=_EMPTY_DICT,
                             allowed=_All_EdgeEl_Dir,
                             required_meta=_EMPTY_DICT,
                             expected_meta=_EMPTY_DICT,
                             type_checked_meta=_EMPTY_DICT,
                             using_hbf_meta=False)


#####################################################################

def _add_by_id_nexson_schema_attributes(container):
    container._using_hbf_meta = True
    container._NexmlEl_Schema = _v1_2_nexml
    container._OtusEl_Schema = _v1_2_Otus
    container._OtuEl_Schema = _v1_2_Otu
    container._TreesEl_Schema = _v1_2_Trees
    container._TreeEl_Schema = _v1_2_Tree
    container._LeafEl_Schema = _v1_2_Leaf
    container._IntNEl_Schema = _v1_2_IntN
    container._EdgeEl_Schema = _v1_2_Edge


def _add_direct_nexson_schema_attributes(container):
    container._using_hbf_meta = True
    container._NexmlEl_Schema = _v1_0_nexml
    container._OtusEl_Schema = _v1_0_Otus
    container._OtuEl_Schema = _v1_0_Otu
    container._TreesEl_Schema = _v1_0_Trees
    container._TreeEl_Schema = _v1_0_Tree
    container._LeafEl_Schema = _v1_0_Leaf
    container._IntNEl_Schema = _v1_0_IntN
    container._NodeEl_Schema = _v1_0_Node
    container._EdgeEl_Schema = _v1_0_Edge


def _add_badgerfish_nexson_schema_attributes(container):
    container._using_hbf_meta = False
    container._NexmlEl_Schema = _v0_0_nexml
    container._OtusEl_Schema = _v0_0_Otus
    container._OtuEl_Schema = _v0_0_Otu
    container._TreesEl_Schema = _v0_0_Trees
    container._TreeEl_Schema = _v0_0_Tree
    container._LeafEl_Schema = _v0_0_Leaf
    container._IntNEl_Schema = _v0_0_IntN
    container._EdgeEl_Schema = _v0_0_Edge
    container._NodeEl_Schema = _v0_0_Node



#!/usr/bin/env python
"""Provides high level wrappers around a Nexson data model blob to
let it be treated as if it were a list of OTUs and a list of trees

Accessors return either references to part of the NexSON or wrappers around
those entities (not copies!)

Weakrefs are used, so the more inclusive containers must be kept in scope
while you are accessing constituents.

Currently converts any NexSON blob to HBF v1.2 data model!
"""
from .syntax import (BY_ID_HONEY_BADGERFISH,
                     convert_nexson_format,
                     detect_nexson_version,
                     get_nexml_el,
                     read_as_json)
from peyutil import is_str_type
import weakref



def otu_iter_nexson_proxy(nexson_proxy, otu_sort=None):
    """otu_sort can be None (not sorted or stable), True (sorted by ID lexigraphically)
    or a key function for a sort function on list of otuIDs

    Note that if there are multiple OTU groups, the NexSON specifies the order of sorting
        of the groups (so the sort argument here only refers to the sorting of OTUs within
        a group)
    """
    nexml_el = nexson_proxy._nexml_el
    og_order = nexml_el['^ot:otusElementOrder']
    ogd = nexml_el['otusById']
    for og_id in og_order:
        og = ogd[og_id]
        if otu_sort is None:
            for k, v in og:
                yield nexson_proxy._create_otu_proxy(k, v)
        else:
            key_list = list(og.keys())
            if otu_sort is True:
                key_list.sort()
            else:
                key_list.sort(key=otu_sort)
            for k in key_list:
                v = og[k]
                yield nexson_proxy._create_otu_proxy(k, v)


def tree_iter_nexson_proxy(nexson_proxy):
    """Iterates over NexsonTreeProxy objects in order determined by the nexson blob"""
    nexml_el = nexson_proxy._nexml_el
    tg_order = nexml_el['^ot:treesElementOrder']
    tgd = nexml_el['treesById']
    for tg_id in tg_order:
        tg = tgd[tg_id]
        tree_order = tg['^ot:treeElementOrder']
        tbid = tg['treeById']
        otus = tg['@otus']
        for k in tree_order:
            v = tbid[k]
            yield nexson_proxy._create_tree_proxy(tree_id=k, tree=v, otus=otus)


def reverse_edge_by_source_dict(ebs, root_id):
    d = {}
    for edge_by_id in ebs.values():
        for edge_id, edge in edge_by_id.items():
            t = edge['@target']
            assert t not in d
            d[t] = (edge_id, edge)
    assert root_id in ebs
    assert root_id not in d
    d[root_id] = (None, None)
    return d


class NexsonProxy(object):
    class NexsonOTUProxy(object):
        def __init__(self, nexson_proxy, otu_id, otu):
            self._nexson_proxy = nexson_proxy
            self._otu_id = otu_id
            self._otu = otu

        def __getitem__(self, key):
            return self.otu[key]

        def __setitem__(self, key, value):
            self.otu[key] = value

        @property
        def ott_id(self):
            return self._otu.get('^ot:ottId')

        @property
        def otu(self):
            return self._otu

        @property
        def _id(self):
            return self._otu_id

        def get(self, key, default=None):
            return self._otu.get(key, default)

        def keys(self, key, default=None):
            return self._otu.get(key, default)

    def __init__(self, filepath='', nexson=None):
        self.filepath = filepath
        if nexson is None:
            if not filepath:
                raise ValueError('Either a filepath or nexson argument must be provided')
            self._nexson = read_as_json(self.filepath)
        else:
            self._nexson = nexson
        v = detect_nexson_version(self._nexson)
        if v != BY_ID_HONEY_BADGERFISH:
            convert_nexson_format(self._nexson, BY_ID_HONEY_BADGERFISH)
        self._nexml_el = get_nexml_el(self._nexson)
        self._otu_cache = {}
        self._tree_cache = {}
        self._wr = None

    def otu_iter(self):
        return iter(otu_iter_nexson_proxy(self))

    def tree_iter(self):
        return iter(tree_iter_nexson_proxy(self))

    def _create_otu_proxy(self, otu_id, otu):
        np = self._otu_cache.get(otu_id)
        if np is None:
            if self._wr is None:
                self._wr = weakref.proxy(self)
            np = NexsonProxy.NexsonOTUProxy(self._wr, otu_id, otu)
            self._otu_cache[otu_id] = np
        return np

    def _create_tree_proxy(self, tree_id, tree, otus):
        np = self._tree_cache.get(tree_id)
        if np is None:
            if self._wr is None:
                self._wr = weakref.proxy(self)
            np = NexsonTreeProxy(tree_id=tree_id, tree=tree, otus=otus, nexson_proxy=self._wr)
            self._tree_cache[tree_id] = np
        return np

    def get_tree(self, tree_id):
        np = self._tree_cache.get(tree_id)
        if np is not None:
            return np
        tgd = self._nexml_el['treesById']
        for tg in tgd.values():
            tbid = tg['treeById']
            if tree_id in tbid:
                otus = tg['@otus']
                tree = tbid[tree_id]
                return self._create_tree_proxy(tree_id=tree_id, tree=tree, otus=otus)
        return None

    def get_otu(self, otu_id):
        np = self._otu_cache.get(otu_id)
        if np is not None:
            return np
        ogd = self._nexml_el['otusById']
        for og in ogd.values():
            o = og['otuById'].get(otu_id)
            if o is not None:
                return self._create_otu_proxy(otu_id=otu_id, otu=o)
        return None


class NexsonTreeProxy(object):
    """Provide more natural operations by wrapping a NexSON 1.2 tree blob and its otus"""

    class NexsonNodeProxy(object):
        def __init__(self, tree, edge_id, edge, node_id=None, node=None):
            self._tree = tree
            self._nexson_tree = tree
            self._node_id = node_id
            self._node = node
            self._edge_id = edge_id
            self._edge = edge
            self._otu = None

        def get(self, key, default=None):
            return self.node.get(key, default)

        def __getitem__(self, key):
            return self.node[key]

        def __setitem__(self, key, value):
            self.node[key] = value

        def keys(self):
            return self.node.keys()

        @property
        def is_leaf(self):
            return self._tree.is_leaf(self.node_id)

        def child_iter(self):
            return self._tree.child_iter(self.node_id)

        @property
        def ott_id(self):
            return self._tree.get_ott_id(self.node)

        @property
        def edge_id(self):
            return self._edge_id

        @property
        def edge(self):
            return self._edge

        @property
        def _id(self):
            return self.node_id

        @property
        def parent(self):
            if self._edge_id is None:
                return None
            par_node_id = self.edge['@source']
            par_edge_id, par_edge = self._tree._find_edge_from_child(par_node_id)
            return self._tree._create_node_proxy_from_edge(edge_id=par_edge_id, edge=par_edge, node_id=par_node_id)

        @property
        def node_id(self):
            if self._node_id is None:
                self._node_id = self._edge['@target']
            return self._node_id

        @property
        def otu(self):
            if self._otu is None:
                otu_id, otu = self._tree._raw_otu_for_node(self.node)
                self._otu = self._tree._nexson_proxy._create_otu_proxy(otu_id=otu_id, otu=otu)
            return self._otu

        @property
        def node(self):
            if self._node is None:
                self._node = self._tree.get_nexson_node(self.node_id)
            return self._node

        def __iter__(self):
            return iter(nexson_tree_preorder_iter(self._tree,
                                                  node=self.node,
                                                  node_id=self.node_id,
                                                  edge_id=self.edge_id,
                                                  edge=self.edge))

        # noinspection PyTypeChecker
        def preorder_iter(self):
            return iter(nexson_tree_preorder_iter(self))

    def __init__(self, tree, tree_id=None, otus=None, nexson_proxy=None):
        self._nexson_proxy = nexson_proxy
        self._nexson_tree = tree
        self._edge_by_source_id = tree['edgeBySourceId']
        self._node_by_source_id = tree['nodeById']
        if is_str_type(otus):
            self._otus_group_id = otus
            self._otus = nexson_proxy._nexml_el['otusById'][otus]['otuById']
        else:
            self._otus = otus
        self._tree_id = tree_id
        # not part of nexson, filled on demand. will be dict of node_id -> (edge_id, edge) pair
        self._edge_by_target = None
        self._wr = None
        self._node_cache = {}

    def get_nexson_node(self, node_id):
        return self._node_by_source_id[node_id]

    def get_node(self, node_id):
        np = self._node_cache.get(node_id)
        if np is not None:
            return np
        edge_id, edge = self._find_edge_from_child(node_id)
        node = self._node_by_source_id[node_id]
        return self._create_node_proxy_from_edge(edge_id, edge, node_id, node)

    def get(self, key, default=None):
        return self._nexson_tree.get(key, default)

    def __getitem__(self, key):
        return self._nexson_tree[key]

    def __setitem__(self, key, value):
        self._nexson_tree[key] = value

    @property
    def edge_by_target(self):
        """Returns a reference to the dict of target node id to (edge_id, edge)"""
        if self._edge_by_target is None:
            self._edge_by_target = reverse_edge_by_source_dict(self._edge_by_source_id,
                                                               self._nexson_tree['^ot:rootNodeId'])
        return self._edge_by_target

    def _find_edge_from_child(self, node_id):
        """Returns (edge_id, edge)"""
        return self.edge_by_target[node_id]

    def _create_node_proxy_from_edge(self, edge_id, edge, node_id=None, node=None):
        np = self._node_cache.get(edge_id)
        if np is None:
            if self._wr is None:
                self._wr = weakref.proxy(self)
            np = NexsonTreeProxy.NexsonNodeProxy(self._wr, edge_id=edge_id, edge=edge, node_id=node_id, node=node)
            self._node_cache[edge_id] = np
            if node_id is not None:
                self._node_cache[node_id] = np
        return np

    def child_iter(self, node_id):
        return nexson_child_iter(self._edge_by_source_id.get(node_id, {}), self)

    def is_leaf(self, node_id):
        return node_id not in self._edge_by_source_id

    def get_ott_id(self, node):
        return self._raw_otu_for_node(node)[1].get('^ot:ottId')

    def _raw_otu_for_node(self, node):
        otu_id = node['@otu']
        # _LOG.debug('otu_id = {}'.format(otu_id))
        return otu_id, self._otus[otu_id]

    def annotate(self, obj, key, value):
        obj[key] = value

    def __iter__(self):
        return iter(nexson_tree_preorder_iter(self))

    def preorder_iter(self):
        return iter(nexson_tree_preorder_iter(self))

    def nodes(self):
        return [i for i in iter(self)]


def nexson_child_iter(edict, nexson_tree_proxy):
    for edge_id, edge in edict.items():
        yield nexson_tree_proxy._create_node_proxy_from_edge(edge_id, edge)


def nexson_tree_preorder_iter(tree_proxy, node_id=None, node=None, edge_id=None, edge=None):
    """Takes a tree in "By ID" NexSON (v1.2).  provides and iterator over:
        NexsonNodeProxy object
    where the edge of the object is the edge connectin the node to the parent.
    The first node will be the root and will have None as it's edge
    """
    tree = tree_proxy._nexson_tree
    ebsid = tree['edgeBySourceId']
    nbid = tree['nodeById']
    if edge_id is not None:
        assert edge is not None
        if node_id is None:
            node_id = edge['@target']
        else:
            assert node_id == edge['@target']
        if node is None:
            node = nbid[node_id]
        else:
            assert node == nbid[node_id]
        yield tree_proxy._create_node_proxy_from_edge(edge_id, edge, node_id=node_id, node=node)
        root_id = node_id
    elif node_id is not None:
        if node is None:
            node = nbid[node_id]
        else:
            assert node == nbid[node_id]
        yield tree_proxy._create_node_proxy_from_edge(None, None, node_id=node_id, node=node)
        root_id = node_id
    else:
        root_id = tree['^ot:rootNodeId']
        root = nbid[root_id]
        yield tree_proxy._create_node_proxy_from_edge(None, None, node_id=root_id, node=root)
    stack = []
    new_stack = [(i['@target'], edge_id, i) for edge_id, i in ebsid[root_id].items()]
    stack.extend(new_stack)
    while stack:
        target_node_id, edge_id, edge = stack.pop()
        yield tree_proxy._create_node_proxy_from_edge(edge_id=edge_id, edge=edge, node_id=target_node_id)
        daughter_edges = ebsid.get(target_node_id)
        if daughter_edges is not None:
            new_stack = [(i['@target'], edge_id, i) for edge_id, i in daughter_edges.items()]
            stack.extend(new_stack)



#!/usr/bin/env python
"""Basic functions and classes which are used by nexson_syntax subpackage,
but do not depend on other parts of nexson.syntax
"""
import re
from peyutil import is_int_type

import logging

_LOG = logging.getLogger(__name__)
# DIRECT_HONEY_BADGERFISH is the closest to BadgerFish
DIRECT_HONEY_BADGERFISH = '1.0.0'
BY_ID_HONEY_BADGERFISH = '1.2.1'
DEFAULT_NEXSON_VERSION = BY_ID_HONEY_BADGERFISH

BADGER_FISH_NEXSON_VERSION = '0.0.0'
NEXML_NEXSON_VERSION = 'nexml'

SUPPORTED_NEXSON_VERSIONS = frozenset([BADGER_FISH_NEXSON_VERSION,
                                       DIRECT_HONEY_BADGERFISH,
                                       BY_ID_HONEY_BADGERFISH])
SUPPORTED_NEXSON_VERSIONS_AND_ALIASES = frozenset([BADGER_FISH_NEXSON_VERSION,
                                                   DIRECT_HONEY_BADGERFISH,
                                                   BY_ID_HONEY_BADGERFISH,
                                                   '0',
                                                   '0.0',
                                                   '1.0',
                                                   '1.2'])
_LITERAL_META_PAT = re.compile(r'.*[:]?LiteralMeta$')
_RESOURCE_META_PAT = re.compile(r'.*[:]?ResourceMeta$')


class NexsonError(Exception):
    def __init__(self, v):
        self.value = v

    def __str__(self):
        return repr(self.value)


def detect_nexson_version(blob):
    """Returns the nexml2json attribute or the default code for badgerfish"""
    n = get_nexml_el(blob)
    assert isinstance(n, dict)
    return n.get('@nexml2json', BADGER_FISH_NEXSON_VERSION)


def get_nexml_el(blob):
    v = blob.get('nexml')
    if v is not None:
        return v
    return blob['nex:nexml']


class NexmlTypeError(Exception):
    def __init__(self, m):
        self.msg = m

    def __str__(self):
        return self.msg


class ConversionConfig(object):
    def __init__(self, output_format, **kwargs):
        self._keys = ['output_format']
        self.output_format = output_format
        for k, v in kwargs.items():
            self.__dict__[k] = v
            self._keys.append(k)

    def items(self):
        for k in self._keys:
            yield k, getattr(self, k)

    def keys(self):
        return list(self._keys)

    def get(self, k, default):
        return getattr(self, k, default)


class NexsonConverter(object):
    def __init__(self, conv_cfg):
        self._conv_cfg = conv_cfg
        for k, v in conv_cfg.items():
            self.__dict__[k] = v
        self.remove_old_structs = conv_cfg.get('remove_old_structs', True)
        self.pristine_if_invalid = conv_cfg.get('pristine_if_invalid', False)


def _index_list_of_values(d, k):
    """Returns d[k] or [d[k]] if the value is not a list"""
    v = d[k]
    if isinstance(v, list):
        return v
    return [v]


def _get_index_list_of_values(d, k, def_value=None):
    """Like _index_list_of_values, but uses get to access and
    returns an empty list if the key is absent.
    Returns d[k] or [d[k]] if the value is not a list"""
    v = d.get(k, def_value)
    if v is None:
        return []
    if isinstance(v, list):
        return v
    return [v]


def _add_value_to_dict_bf(d, k, v):
    """Adds the `k`->`v` mapping to `d`, but if a previous element exists it changes
    the value of for the key to list.

    This is used in the BadgerFish mapping convention.

    This is a simple multi-dict that is only suitable when you know that you'll never
    store a list or `None` as a value in the dict.
    """
    prev = d.get(k)
    if prev is None:
        d[k] = v
    elif isinstance(prev, list):
        if isinstance(v, list):
            prev.extend(v)
        else:
            prev.append(v)
    else:
        if isinstance(v, list):
            x = [prev]
            x.extend(v)
            d[k] = x
        else:
            d[k] = [prev, v]


def _add_uniq_value_to_dict_bf(d, k, v):
    """Like _add_value_to_dict_bf but will not add v if another
    element in under key `k` has the same value.
    """
    prev = d.get(k)
    if prev is None:
        d[k] = v
    elif isinstance(prev, list):
        if not isinstance(v, list):
            v = [v]
        for sel in v:
            found = False
            for el in prev:
                if el == sel:
                    found = True
                    break
            if not found:
                prev.append(sel)
    else:
        if isinstance(v, list):
            prev = [prev]
            for sel in v:
                found = False
                for el in prev:
                    if el == sel:
                        found = True
                        break
                if not found:
                    prev.append(sel)
            if len(prev) > 1:
                d[k] = prev
        elif prev != v:
            d[k] = [prev, v]


_is_badgerfish_version = lambda x: x.startswith('0.')
_is_direct_hbf = lambda x: x.startswith('1.0.')
_is_by_id_hbf = lambda x: x.startswith('1.2')
_is_supported_nexson_vers = lambda x: x in SUPPORTED_NEXSON_VERSIONS_AND_ALIASES


def _debug_dump_dom(el):
    """Debugging helper. Prints out `el` contents."""
    import xml.dom.minidom
    s = [el.nodeName]
    att_container = el.attributes
    for i in range(att_container.length):
        attr = att_container.item(i)
        s.append('  @{a}="{v}"'.format(a=attr.name, v=attr.value))
    for c in el.childNodes:
        if c.nodeType == xml.dom.minidom.Node.TEXT_NODE:
            s.append('  {a} type="TEXT" data="{d}"'.format(a=c.nodeName, d=c.data))
        else:
            s.append('  {a} child'.format(a=c.nodeName))
    return '\n'.join(s)


def _cull_redundant_about(obj):
    """Removes the @about key from the `obj` dict if that value refers to the
    dict's '@id'
    """
    about_val = obj.get('@about')
    if about_val:
        id_val = obj.get('@id')
        if id_val and (('#' + id_val) == about_val):
            del obj['@about']


def _add_redundant_about(obj):
    id_val = obj.get('@id')
    if id_val and ('@about' not in obj):
        obj['@about'] = ('#' + id_val)


def _coerce_literal_val_to_primitive(datatype, str_val):
    _TYPE_ERROR_MSG_FORMAT = 'Expected meta property to have type {t}, but found "{v}"'
    if datatype == 'xsd:string':
        return str_val
    if datatype in frozenset(['xsd:int', 'xsd:integer', 'xsd:long']):
        try:
            return int(str_val)
        except:
            raise NexmlTypeError(_TYPE_ERROR_MSG_FORMAT.format(t=datatype, v=str_val))
    elif datatype == frozenset(['xsd:float', 'xsd:double']):
        try:
            return float(str_val)
        except:
            raise NexmlTypeError(_TYPE_ERROR_MSG_FORMAT.format(t=datatype, v=str_val))
    elif datatype == 'xsd:boolean':
        if str_val.lower() in frozenset(['1', 'true']):
            return True
        elif str_val.lower() in frozenset(['0', 'false']):
            return False
        else:
            raise NexmlTypeError(_TYPE_ERROR_MSG_FORMAT.format(t=datatype, v=str_val))
    else:
        _LOG.debug('unknown xsi:type "%s"', datatype)
        return None  # We'll fall through to here when we encounter types we do not recognize


def _python_instance_to_nexml_meta_datatype(v):
    """Returns 'xsd:string' or a more specific type for a <meta datatype="XYZ"...
    syntax using introspection.
    """
    if isinstance(v, bool):
        return 'xsd:boolean'
    if is_int_type(v):
        return 'xsd:int'
    if isinstance(v, float):
        return 'xsd:float'
    return 'xsd:string'


# noinspection PyTypeChecker
def _convert_hbf_meta_val_for_xml(key, val):
    """Convert to a BadgerFish-style dict for addition to a dict suitable for
    addition to XML tree or for v1.0 to v0.0 conversion."""
    if isinstance(val, list):
        return [_convert_hbf_meta_val_for_xml(key, i) for i in val]
    is_literal = True
    content = None
    if isinstance(val, dict):
        ret = val
        if '@href' in val:
            is_literal = False
        else:
            content = val.get('$')
            if isinstance(content, dict) and _contains_hbf_meta_keys(val):
                is_literal = False
    else:
        ret = {}
        content = val
    if is_literal:
        ret.setdefault('@xsi:type', 'nex:LiteralMeta')
        ret.setdefault('@property', key)
        if content is not None:
            ret.setdefault('@datatype', _python_instance_to_nexml_meta_datatype(content))
        if ret is not val:
            ret['$'] = content
    else:
        ret.setdefault('@xsi:type', 'nex:ResourceMeta')
        ret.setdefault('@rel', key)
    return ret


def get_bf_meta_value(d):
    v = d.get('$')
    if v is not None:
        return v
    v = d.get('@content')
    if v is not None:
        return v
    return d.get('@href')


def _contains_hbf_meta_keys(d):
    for k in d.keys():
        if k.startswith('^'):
            return True
    return False


def extract_meta(x):
    try:
        return get_bf_meta_value(x)
    except:
        return None


def find_val_for_first_bf_l_meta(d, prop_name):
    """Returns the $ value of the first meta element with
    the @property that matches @prop_name (or None).
    """
    m_list = d.get('meta')
    if not m_list:
        return None
    if not isinstance(m_list, list):
        m_list = [m_list]
    for m_el in m_list:
        if m_el.get('@property') == prop_name:
            return extract_meta(m_el)
    return None


def find_val_for_first_bf_r_meta(d, prop_name):
    """Returns the $ value of the first meta element with
    the @rel that matches @prop_name (or None).
    """
    m_list = d.get('meta')
    if not m_list:
        return None
    if not isinstance(m_list, list):
        m_list = [m_list]
    for m_el in m_list:
        if m_el.get('@rel') == prop_name:
            return extract_meta(m_el)
    return None


def find_val_for_first_hbf_l_meta(d, prop_name):
    p = '^' + prop_name
    return d.get(p)


def find_val_literal_meta_first(d, prop_name, version):
    if _is_badgerfish_version(version):
        return find_val_for_first_bf_l_meta(d, prop_name)
    p = '^' + prop_name
    return d.get(p)


def find_nested_meta_first_bf(d, prop_name):
    """Returns the $ value of the first meta element with
    the @property that matches @prop_name (or None).
    """
    m_list = d.get('meta')
    if not m_list:
        return None
    if not isinstance(m_list, list):
        m_list = [m_list]
    for m_el in m_list:
        if m_el.get('@property') == prop_name or m_el.get('@rel') == prop_name:
            return m_el
    return None


def find_nested_meta_first(d, prop_name, version):
    """Returns obj. for badgerfish and val for hbf. Appropriate for nested literals"""
    if _is_badgerfish_version(version):
        return find_nested_meta_first_bf(d, prop_name)
    p = '^' + prop_name
    return d.get(p)


def find_val_resource_meta_first(d, prop_name, version):
    if _is_badgerfish_version(version):
        return find_val_for_first_bf_r_meta(d, prop_name)
    p = '^' + prop_name
    return d.get(p)


def add_literal_meta(obj, prop_name, value, version):
    if _is_badgerfish_version(version):
        m = obj.setdefault('meta', [])
        if not isinstance(m, list):
            m = [m]
            obj['meta'] = m
        d = {'$': value,
             '@property': prop_name,
             '@xsi:type': 'nex:LiteralMeta'}
        m.append(d)
        return d
    else:
        k = '^' + prop_name
        _add_value_to_dict_bf(obj, k, value)
        return value


def delete_first_literal_meta(obj, prop_name, version):
    if _is_badgerfish_version(version):
        m = obj.setdefault('meta', [])
        if not isinstance(m, list):
            m = [m]
            obj['meta'] = m
        ind = None
        for n, el in enumerate(m):
            if el.get('@property') == prop_name:
                ind = n
                break
        if ind is not None:
            m.pop(ind)
        if len(m) == 0:
            del obj['meta']
    else:
        k = '^' + prop_name
        if k in obj:
            del obj[k]


def _simplify_object_by_id_del(o):
    if isinstance(o, list):
        return [_simplify_object_by_id_del(i) for i in o]
    if not isinstance(o, dict):
        return o
    if '@id' in o:
        nk = len(o.keys())
        if nk < 3:
            if nk == 1:
                return None
            if '$' in o:
                return o['$']
            if '@href' in o:
                del o['@id']
    return o


def _simplify_all_meta_by_id_del(el):
    to_del = []
    for tag in el.keys():
        if tag.startswith('^'):
            o = el[tag]
            v = _simplify_object_by_id_del(o)
            if v is None:
                to_del.append(tag)
            elif v is not o:
                el[tag] = v
    for tag in to_del:
        del el[tag]



#!/usr/bin/env python
"""Simple inspection of NexSON blob
 Functions here should:
    - work on any nexson version that the nexon package supports and,
    - avoid any coercion of the NexSON to another version.
"""
from __future__ import absolute_import, print_function, division
from nexson.syntax import detect_nexson_version, get_nexml_el, _is_by_id_hbf
import logging

_LOG = logging.getLogger(__name__)


def count_num_trees(nexson, nexson_version=None):
    """Returns the number of trees summed across all tree
    groups.
    """
    if nexson_version is None:
        nexson_version = detect_nexson_version(nexson)
    nex = get_nexml_el(nexson)
    num_trees_by_group = []
    if _is_by_id_hbf(nexson_version):
        for tree_group in nex.get('treesById', {}).values():
            nt = len(tree_group.get('treeById', {}))
            num_trees_by_group.append(nt)
    else:
        trees_group = nex.get('trees', [])
        if isinstance(trees_group, dict):
            trees_group = [trees_group]
        for tree_group in trees_group:
            t = tree_group.get('tree')
            if isinstance(t, list):
                nt = len(t)
            else:
                nt = 1
            num_trees_by_group.append(nt)
    return sum(num_trees_by_group)



#!/usr/bin/env python
"""NexsonValidationAdaptor class.
"""
from nexson.validation._badgerfish_validation import BadgerFishValidationAdaptor
from nexson.validation._by_id_validation import ByIdHBFValidationAdaptor
from nexson.syntax import detect_nexson_version
from nexson.syntax.helper import (find_val_for_first_hbf_l_meta,
                             DIRECT_HONEY_BADGERFISH,
                             _is_badgerfish_version,
                             _is_by_id_hbf,
                             _is_direct_hbf)
import logging

_LOG = logging.getLogger(__name__)


class DirectHBFValidationAdaptor(BadgerFishValidationAdaptor):
    def __init__(self, obj, logger, **kwargs):
        self._find_first_literal_meta = find_val_for_first_hbf_l_meta
        self._syntax_version = DIRECT_HONEY_BADGERFISH
        BadgerFishValidationAdaptor.__init__(self, obj, logger, **kwargs)


def  create_validation_adaptor(obj, logger, **kwargs):
    try:
        nexson_version = detect_nexson_version(obj)
    except:
        return BadgerFishValidationAdaptor(obj, logger, **kwargs)
    if _is_by_id_hbf(nexson_version):
        # _LOG.debug('validating as ById...')
        return ByIdHBFValidationAdaptor(obj, logger, **kwargs)
    elif _is_badgerfish_version(nexson_version):
        # _LOG.debug('validating as BadgerFish...')
        return BadgerFishValidationAdaptor(obj, logger, **kwargs)
    elif _is_direct_hbf(nexson_version):
        # _LOG.debug('validating as DirectHBF...')
        return DirectHBFValidationAdaptor(obj, logger, **kwargs)
    raise NotImplementedError('nexml2json version {v}'.format(v=nexson_version))



#!/usr/bin/env python
"""Badgerfish2DirectNexson class"""
from .helper import (_RESOURCE_META_PAT,
                     _LITERAL_META_PAT,
                    DIRECT_HONEY_BADGERFISH,
                     _add_value_to_dict_bf,
                     _coerce_literal_val_to_primitive,
                     _cull_redundant_about,
                     _get_index_list_of_values,
                     get_nexml_el,
                     NexsonConverter)
from peyutil import is_str_type
import logging

_LOG = logging.getLogger(__name__)

_SUPPRESSED_LITERAL = frozenset(['@datatype', '@property', '@xsi:type', '$'])
_SUPPRESSED_RESOURCE = frozenset(['@rel', '@property', '@xsi:type'])


class Badgerfish2DirectNexson(NexsonConverter):
    """Conversion of the direct Badgerfish and Phylografter JSON
    to the direct form of honeybadgerfish JSON (v 1.0)
    This is a dict-to-dict in-place conversion. No serialization is included.
    """

    def __init__(self, conv_cfg):
        NexsonConverter.__init__(self, conv_cfg)
        self._from_phylografter = conv_cfg.get('_from_phylografter', True)
        if self._from_phylografter:
            self._workaround_phylografter_rel_bug = True  # pg uses "property" instead of "rel"
            self._add_tree_xsi_type = True
            self._coercing_literals = True  # phylografter uses "true" for isLeaf
        else:
            self._workaround_phylografter_rel_bug = False
            self._add_tree_xsi_type = False

    def _transform_literal_meta(self, lit_bf_meta):
        dt = lit_bf_meta.get('@datatype')
        content = lit_bf_meta.get('$')
        att_key = lit_bf_meta['@property']
        full_obj = {}
        for k in lit_bf_meta.keys():
            if k not in _SUPPRESSED_LITERAL:
                full_obj[k] = lit_bf_meta[k]
        # Coercion should not be needed for json->json
        if dt and self._coercing_literals:
            if is_str_type(content):
                content = _coerce_literal_val_to_primitive(dt, content)
        att_key = '^' + att_key
        if full_obj:
            if content:
                full_obj['$'] = content
            _cull_redundant_about(full_obj)
            return att_key, full_obj
        return att_key, content

    def _transform_resource_meta(self, res_bf_meta):
        try:
            att_key = res_bf_meta['@rel']
        except KeyError:
            if self._workaround_phylografter_rel_bug:
                att_key = res_bf_meta['@property']
            else:
                raise
        full_obj = {}
        for k in res_bf_meta.keys():
            if k not in _SUPPRESSED_RESOURCE:
                full_obj[k] = res_bf_meta[k]
        att_key = '^' + att_key
        assert full_obj
        _cull_redundant_about(full_obj)
        return att_key, full_obj

    def _recursive_convert_list(self, obj):
        for el in obj:
            if isinstance(el, dict):
                self._recursive_convert_dict(el)

    def _recursive_convert_dict(self, obj):
        _cull_redundant_about(obj)  # rule 10...
        meta_list = _get_index_list_of_values(obj, 'meta')
        to_inject = {}
        for meta in meta_list:
            xt = meta['@xsi:type']
            if _RESOURCE_META_PAT.match(xt):
                mk, mv = self._transform_resource_meta(meta)
            else:
                assert _LITERAL_META_PAT.match(xt)
                mk, mv = self._transform_literal_meta(meta)
            _add_value_to_dict_bf(to_inject, mk, mv)
        if ('meta' in obj) and self.remove_old_structs:
            del obj['meta']
        for k, v in to_inject.items():
            _add_value_to_dict_bf(obj, k, v)
        for k, v in obj.items():
            if isinstance(v, dict):
                self._recursive_convert_dict(v)
            elif isinstance(v, list):
                self._recursive_convert_list(v)

    def _dict_to_list_of_dicts(self, obj, tag, child_tag=None, grand_child_tag=None):
        el = obj.get(tag)
        if el:
            if not isinstance(el, list):
                as_list = [el]
                if child_tag is None:
                    obj[tag] = as_list
                    return
            else:
                as_list = el
            if child_tag:
                for sub in as_list:
                    self._dict_to_list_of_dicts(sub, tag=child_tag, child_tag=grand_child_tag)

    def convert(self, obj):
        """Takes a dict corresponding to the honeybadgerfish JSON blob of the 1.0.* type and
        converts it to BY_ID_HONEY_BADGERFISH version. The object is modified in place
        and returned.
        """
        if self.pristine_if_invalid:
            raise NotImplementedError('pristine_if_invalid option is not supported yet')

        nex = get_nexml_el(obj)
        assert nex
        self._recursive_convert_dict(nex)
        # pluralization simplifications in hbf:
        # convert dicts to lists for the primary datastructures...
        self._dict_to_list_of_dicts(nex, 'otus')
        self._dict_to_list_of_dicts(nex, 'otus', 'otu')
        self._dict_to_list_of_dicts(nex, 'trees')
        self._dict_to_list_of_dicts(nex, 'trees', 'tree')
        self._dict_to_list_of_dicts(nex, 'trees', 'tree', 'node')
        self._dict_to_list_of_dicts(nex, 'trees', 'tree', 'edge')
        if self._add_tree_xsi_type:
            for tb in nex.get('trees', []):
                for t in tb.get('tree', []):
                    t.setdefault('@xsi:type', 'nex:FloatTree')
        nex['@nexml2json'] = str(DIRECT_HONEY_BADGERFISH)
        return obj



#!/usr/bin/env python
from .helper import _NEXEL, error_return
from .schema import check_raw_dict
from .err_generator import (gen_MissingCrucialContentWarning,
                            gen_MultipleRootsWarning,
                            gen_NodeWithMultipleParents,
                            gen_NoRootWarning,
                            gen_ReferencedIDNotFoundWarning,
                            gen_RepeatedIDWarning,
                            gen_RepeatedOTUWarning,
                            gen_UnreachableNodeWarning,
                            gen_WrongValueTypeWarning)
from peyutil import is_str_type
from ._validation_base import NexsonValidationAdaptor
# noinspection PyUnresolvedReferences,PyUnresolvedReferences
from nexson.syntax.helper import BY_ID_HONEY_BADGERFISH
import logging

_LOG = logging.getLogger(__name__)


class ByIdHBFValidationAdaptor(NexsonValidationAdaptor):
    def __init__(self, obj, logger, **kwargs):
        self._syntax_version = BY_ID_HONEY_BADGERFISH
        NexsonValidationAdaptor.__init__(self, obj, logger, **kwargs)

    def _post_key_check_validate_otus_obj(self, og_nex_id, otus_group, vc):
        otu_obj = otus_group.get('otuById', {})
        if not isinstance(otu_obj, dict):
            self._error_event(_NEXEL.OTUS,
                              obj=otus_group,
                              err_type=gen_WrongValueTypeWarning,
                              anc=vc.anc_list,
                              obj_nex_id=og_nex_id,
                              key_list=['otuById'])
            return error_return('no "otuById" in otus group')
        self._otu_group_by_id[og_nex_id] = otus_group
        vc.push_context(_NEXEL.OTU, (otus_group, og_nex_id))
        try:
            not_dict_otu = []
            ndo_id_list = []
            otu_id_obj_list = []
            for id_obj_pair in otu_obj.items():
                if not isinstance(id_obj_pair[1], dict):
                    r = check_raw_dict(id_obj_pair[1], otus_group, None, vc)
                    assert r[0] is False
                    t = r[1]
                    not_dict_otu.append(t)
                    ndo_id_list.append(id_obj_pair[0])
                else:
                    otu_id_obj_list.append(id_obj_pair)
            if not_dict_otu:
                self._error_event(_NEXEL.OTU,
                                  obj=otu_obj,
                                  err_type=gen_WrongValueTypeWarning,
                                  anc=vc.anc_list,
                                  obj_nex_id=ndo_id_list,
                                  key_val_type_list=[not_dict_otu])
                return error_return('otu is wrong type')
            r = self._validate_otu_list(otu_id_obj_list, vc)
            if r:
                self._otu_by_otug[og_nex_id] = otu_obj
            return r
        finally:
            vc.pop_context()

    def _post_key_check_validate_tree_group(self, tg_nex_id, tree_group_obj, vc):
        tree_id_order = tree_group_obj.get('^ot:treeElementOrder', [])
        tree_by_id = tree_group_obj.get('treeById', {})
        if (not isinstance(tree_by_id, dict)) \
                or (not isinstance(tree_id_order, list)):
            self._error_event(_NEXEL.TREES,
                              obj=tree_group_obj,
                              err_type=gen_WrongValueTypeWarning,
                              anc=vc.anc_list,
                              obj_nex_id=tg_nex_id,
                              key_list=['treeById', '^ot:treeElementOrder'])
            return error_return('no "treeById" in trees group')
        if ((not tree_by_id) and tree_id_order) or ((not tree_id_order) and tree_by_id):
            self._error_event(_NEXEL.TREES,
                              obj=tree_group_obj,
                              err_type=gen_MissingCrucialContentWarning,
                              anc=vc.anc_list,
                              obj_nex_id=tg_nex_id,
                              key_list=['treeById', '^ot:treeElementOrder'])
            return error_return('no "treeById" in trees group')
        otus_el = tree_group_obj.get('@otus')
        if otus_el not in self._otu_group_by_id:
            self._error_event(_NEXEL.TREES,
                              obj=tree_group_obj,
                              err_type=gen_ReferencedIDNotFoundWarning,
                              anc=vc.anc_list,
                              obj_nex_id=tg_nex_id,
                              key_list=[otus_el])
            return error_return('no "@otus" in trees group')
        for t_nex_id, tree_obj in tree_by_id.items():
            vc.push_context(_NEXEL.TREE, (tree_group_obj, tg_nex_id))
            try:
                if not self._validate_tree(t_nex_id, tree_obj, vc, otus_group_id=otus_el):
                    return False
            finally:
                vc.pop_context()
        return True

    def _post_key_check_validate_tree(self,
                                      tree_nex_id,
                                      tree_obj,
                                      vc,
                                      otus_group_id=None):
        # pylint: disable=R0914
        node_by_id = tree_obj.get('nodeById')
        edge_by_source = tree_obj.get('edgeBySourceId')
        root_node_id = tree_obj.get('^ot:rootNodeId')
        if (not node_by_id) or (not isinstance(node_by_id, dict)):
            self._error_event(_NEXEL.TREE,
                              obj=tree_obj,
                              err_type=gen_MissingCrucialContentWarning,
                              anc=vc.anc_list,
                              obj_nex_id=tree_nex_id,
                              key_list=['nodeById', ])
            return error_return('no "nodeById" in tree')
        if (not edge_by_source) or (not isinstance(edge_by_source, dict)):
            self._error_event(_NEXEL.TREE,
                              obj=tree_obj,
                              err_type=gen_MissingCrucialContentWarning,
                              anc=vc.anc_list,
                              obj_nex_id=tree_nex_id,
                              key_list=['edgeBySourceId', ])
            return error_return('no "edgeBySourceId" in tree')
        if not is_str_type(root_node_id):
            self._error_event(_NEXEL.TREE,
                              obj=tree_obj,
                              err_type=gen_MissingCrucialContentWarning,
                              anc=vc.anc_list,
                              obj_nex_id=tree_nex_id,
                              key_list=['^ot:rootNodeId', ])
            return error_return('no "^ot:rootNodeId" in tree')
        edge_dict = {}
        edge_by_target = {}
        if otus_group_id is None:
            tree_group = vc.anc_list[-1][1]
            otus_group_id = tree_group.get('@otus')

        bad_node_ref = []
        repeated_edge_id = []
        missing_target = []
        repeated_target = []
        reachable_nodes = set()
        for par_node_id, edge_by_id in edge_by_source.items():
            if par_node_id not in node_by_id:
                bad_node_ref.append(par_node_id)
            else:
                reachable_nodes.add(par_node_id)
                for edge_id, edge in edge_by_id.items():
                    if edge_id in edge_dict:
                        repeated_edge_id.append(edge_id)
                    else:
                        edge_dict[edge_id] = edge
                        try:
                            t = edge.get('@target')
                        except:
                            t = None
                        if t is None:
                            missing_target.append(edge_id)
                        elif t in edge_by_target:
                            repeated_target.append(t)
                        else:
                            edge_by_target[t] = edge
                            reachable_nodes.add(t)
        node_set = set(node_by_id.keys())
        if node_set != reachable_nodes:
            unreachable = list(node_set - reachable_nodes)
            unreachable.sort()
            not_in_node_by_id = list(reachable_nodes - node_set)
            not_in_node_by_id.sort()
            if unreachable:
                self._error_event(_NEXEL.TREE,
                                  obj=tree_obj,
                                  err_type=gen_UnreachableNodeWarning,
                                  anc=vc.anc_list,
                                  obj_nex_id=tree_nex_id,
                                  key_list=unreachable)
                return error_return('unreachable node in tree tree')
            if not_in_node_by_id:
                self._error_event(_NEXEL.TREE,
                                  obj=tree_obj,
                                  err_type=gen_ReferencedIDNotFoundWarning,
                                  anc=vc.anc_list,
                                  obj_nex_id=tree_nex_id,
                                  key_list=not_in_node_by_id)
                return error_return('referenced node id not in "nodeById" in tree')
        if bad_node_ref:
            bad_node_ref.sort()
            self._error_event(_NEXEL.TREE,
                              obj=tree_obj,
                              err_type=gen_ReferencedIDNotFoundWarning,
                              anc=vc.anc_list,
                              obj_nex_id=tree_nex_id,
                              key_list=bad_node_ref)
            return error_return('referenced parent node not in "nodeById" in tree')
        if missing_target:
            missing_target.sort()
            self._error_event(_NEXEL.TREE,
                              obj=tree_obj,
                              err_type=gen_ReferencedIDNotFoundWarning,
                              anc=vc.anc_list,
                              obj_nex_id=tree_nex_id,
                              key_list=missing_target)
            return error_return('no "@target" in edge')
        if repeated_target:
            repeated_target.sort()
            self._error_event(_NEXEL.TREE,
                              obj=tree_obj,
                              err_type=gen_NodeWithMultipleParents,
                              anc=vc.anc_list,
                              obj_nex_id=tree_nex_id,
                              node_id_list=repeated_target)
            return error_return('same node used as "@target" for different edges')
        if repeated_edge_id:
            repeated_edge_id.sort()
            self._error_event(_NEXEL.EDGE,
                              obj=tree_obj,
                              err_type=gen_RepeatedIDWarning,
                              anc=vc.anc_list,
                              obj_nex_id=tree_nex_id,
                              key_list=repeated_edge_id)
            return error_return('edge "@id" repeated')
        node_set = set(edge_by_target.keys())
        internal_node_set = set(edge_by_source.keys())
        leaf_set = node_set - internal_node_set
        leaves = [(i, node_by_id[i]) for i in leaf_set]
        vc.push_context(_NEXEL.LEAF_NODE, (tree_obj, tree_nex_id))
        try:
            if not self._validate_leaf_list(leaves, vc):
                return False
        finally:
            vc.pop_context()
        internal_nodes = [(i, node_by_id[i]) for i in internal_node_set]
        with_at_root_prop = {}
        for nid, n_obj in internal_nodes:
            if n_obj.get('@root'):
                with_at_root_prop[nid] = n_obj
        if len(with_at_root_prop) > 1:
            self._error_event(_NEXEL.TREE,
                              obj=tree_obj,
                              err_type=gen_MultipleRootsWarning,
                              anc=vc.anc_list,
                              obj_nex_id=tree_nex_id,
                              node_id_list=list(with_at_root_prop.keys()))
            return error_return('multiple "@root" nodes')
        if len(with_at_root_prop) == 0:
            self._error_event(_NEXEL.TREE,
                              obj=tree_obj,
                              err_type=gen_NoRootWarning,
                              anc=vc.anc_list,
                              obj_nex_id=tree_nex_id)
            return error_return('No node labelled as "@root"')
        if root_node_id not in with_at_root_prop:
            self._error_event(_NEXEL.TREE,
                              obj=tree_obj,
                              err_type=gen_MultipleRootsWarning,
                              anc=vc.anc_list,
                              obj_nex_id=tree_nex_id,
                              node_id_list=list(with_at_root_prop.keys()) + [root_node_id])
            return error_return('root node not labelled as root')
        vc.push_context(_NEXEL.INTERNAL_NODE, (tree_obj, tree_nex_id))
        try:
            if not self._validate_internal_node_list(internal_nodes, vc):
                return False
        finally:
            vc.pop_context()
        edges = [i for i in edge_dict.items()]
        vc.push_context(_NEXEL.EDGE, (tree_obj, tree_nex_id))
        try:
            if not self._validate_edge_list(edges, vc):
                return False
        finally:
            vc.pop_context()
        otuid2leaf = {}
        for nd_id, nd in leaves:
            otuid = nd['@otu']
            if otuid in otuid2leaf:
                vc.push_context(_NEXEL.LEAF_NODE, (tree_obj, tree_nex_id))
                try:
                    self._error_event(_NEXEL.LEAF_NODE,
                                      obj=nd,
                                      err_type=gen_RepeatedOTUWarning,
                                      anc=vc.anc_list,
                                      obj_nex_id=nd_id,
                                      key_list=[otuid])
                    return error_return('Repeated "@otu" id')
                finally:
                    vc.pop_context()
            otuid2leaf[otuid] = nd_id
        self._detect_multilabelled_tree(otus_group_id=otus_group_id,
                                        tree_id=tree_nex_id,
                                        otuid2leaf=otuid2leaf)
        og = self._otu_group_by_id[otus_group_id]['otuById']
        return self._validate_otu_key_if_present(node_by_id.items(), og, vc)

    def _post_key_check_validate_nexml_obj(self, nex_obj, obj_nex_id, vc):
        otus = nex_obj.get('otusById', {})
        otus_order_list = nex_obj.get('^ot:otusElementOrder', [])
        if not isinstance(otus, dict) or (otus_order_list and (not otus)):
            self._error_event(_NEXEL.NEXML,
                              obj=nex_obj,
                              err_type=gen_WrongValueTypeWarning,
                              anc=vc.anc_list,
                              obj_nex_id=obj_nex_id,
                              key_list=['otusById'])
            return error_return('Missing "otusById"')
        if (not isinstance(otus_order_list, list)) or ((not otus_order_list) and otus):
            self._error_event(_NEXEL.NEXML,
                              obj=nex_obj,
                              err_type=gen_WrongValueTypeWarning,
                              anc=vc.anc_list,
                              obj_nex_id=obj_nex_id,
                              key_list=['^ot:otusElementOrder'])
            return error_return('Missing "^ot:otusElementOrder"')
        otus_group_list = []
        missing_ogid = []
        not_dict_og = []
        for ogid in otus_order_list:
            og = otus.get(ogid)
            if og is None:
                missing_ogid.append(ogid)
            elif not isinstance(og, dict):
                r = check_raw_dict(og, otus, ogid, vc)
                assert r[0] is False
                t = r[1]
                not_dict_og.append(t)
            else:
                otus_group_list.append((ogid, og))
        if missing_ogid:
            self._error_event(_NEXEL.NEXML,
                              obj=nex_obj,
                              err_type=gen_ReferencedIDNotFoundWarning,
                              anc=vc.anc_list,
                              obj_nex_id=obj_nex_id,
                              key_list=missing_ogid)
            return error_return('Missing "@id" for otus group')
        vc.push_context(_NEXEL.OTUS, (nex_obj, obj_nex_id))
        try:
            if not_dict_og:
                self._error_event(_NEXEL.OTUS,
                                  obj=otus,
                                  err_type=gen_WrongValueTypeWarning,
                                  anc=vc.anc_list,
                                  obj_nex_id=None,
                                  key_val_type_list=[not_dict_og])
                return error_return('Otus objects of the wrong type of object')
            if not self._validate_otus_group_list(otus_group_list, vc):
                return False
        finally:
            vc.pop_context()
        # and now the trees...
        trees = nex_obj.get('treesById')
        trees_order_list = nex_obj.get('^ot:treesElementOrder')
        if (not isinstance(trees, dict)) or ((not trees) and trees_order_list):
            self._error_event(_NEXEL.NEXML,
                              obj=nex_obj,
                              err_type=gen_MissingCrucialContentWarning,
                              anc=vc.anc_list,
                              obj_nex_id=obj_nex_id,
                              key_list=['treesById'])
            return error_return('Missing "treesById"')
        if (not isinstance(trees_order_list, list)) or ((not trees_order_list) and trees):
            self._error_event(_NEXEL.NEXML,
                              obj=nex_obj,
                              err_type=gen_MissingCrucialContentWarning,
                              anc=vc.anc_list,
                              obj_nex_id=obj_nex_id,
                              key_list=['^ot:treesElementOrder'])
            return error_return('Missing "^ot:treesElementOrder"')
        trees_group_list = []
        missing_tgid = []
        not_dict_tg = []
        for tgid in trees_order_list:
            tg = trees.get(tgid)
            if tg is None:
                missing_tgid.append(tgid)
            elif not isinstance(tg, dict):
                r = check_raw_dict(tg, trees, tgid, vc)
                assert r[0] is False
                t = r[1]
                not_dict_tg.append(t)
            else:
                trees_group_list.append((tgid, tg))
        if missing_tgid:
            self._error_event(_NEXEL.NEXML,
                              obj=nex_obj,
                              err_type=gen_ReferencedIDNotFoundWarning,
                              anc=vc.anc_list,
                              obj_nex_id=obj_nex_id,
                              key_list=missing_tgid)
            return error_return('Missing trees group id')
        vc.push_context(_NEXEL.TREES, (nex_obj, obj_nex_id))
        try:
            if not_dict_tg:
                self._error_event(_NEXEL.TREES,
                                  obj=trees,
                                  err_type=gen_WrongValueTypeWarning,
                                  anc=vc.anc_list,
                                  obj_nex_id=None,
                                  key_val_type_list=[not_dict_tg])
                return error_return('Trees element of the wrong type')
            if not self._validate_trees_group_list(trees_group_list, vc):
                return False
        finally:
            vc.pop_context()
        if not nex_obj.get('ot:notIntendedForSynthesis'):
            cs = nex_obj.get('ot:candidateTreeForSynthesis')
            if cs:
                if not isinstance(cs, list):
                    tree_list = [cs]
                else:
                    tree_list = cs
            else:
                tree_list = []
                for tg in trees.values():
                    tbi = tg.get('treeById', {})
                    tree_list.extend(tbi.keys())
            self._generate_ott_warnings(otus, tree_list, (nex_obj, obj_nex_id), vc)
        return True



#!/usr/bin/env python
"""Functions for validating NexSON.
"""
from nexson.validation.adaptor import create_validation_adaptor
from nexson.validation.logger import FilteringLogger, ValidationLogger
from nexson.validation.warning_codes import NexsonWarningCodes
from nexson.syntax import NexsonError

def validate_nexson(obj,
                    warning_codes_to_skip=None,
                    retain_deprecated=True,
                    **kwargs):
    """Takes a NexSON `obj`. Returns a `(validatation_log, adaptor)` pair.

    `warning_codes_to_skip` is None or an iterable collection of integers
        that correspond to facets of NexsonWarningCodes. Warnings associated
        with any code in the list will be suppressed.
    `retain_deprecated` if False, then `obj` may be modified to replace
        deprecated constructs with new syntax. If it is True, the `obj` will
        not be modified.

    `validatation_log` will be an instance of a `DefaultRichLogger` subclass.
        it will hold warning and error messages.
    `adaptor` will be an instance of `NexsonValidationAdaptor`.
        it holds a reference to `obj` and the bookkeepping data necessary to attach
        the log message to `obj` if
    """
    if warning_codes_to_skip:
        v = FilteringLogger(codes_to_skip=list(warning_codes_to_skip), store_messages=True)
    else:
        v = ValidationLogger(store_messages=True)
    v.retain_deprecated = retain_deprecated
    n = create_validation_adaptor(obj, v, **kwargs)
    return v, n


def ot_validate(nexson, **kwargs):
    """Returns three objects:
        an annotation dict (NexSON formmatted),
        the validation_log object created when NexSON validation was performed, and
        the object of class NexSON which was created from nexson. This object may
            alias parts of the nexson dict that is passed in as an argument.

    Currently the only kwargs used is 'max_num_trees_per_study'
    """
    # stub function for hooking into NexSON validation
    codes_to_skip = [NexsonWarningCodes.UNVALIDATED_ANNOTATION]  # pylint: disable=E1101
    v_log, adaptor_obj = validate_nexson(nexson, codes_to_skip, **kwargs)
    annotation = v_log.prepare_annotation(author_name='api.opentreeoflife.org/validate',
                                          description='Open Tree NexSON validation')
    return annotation, v_log, adaptor_obj

from ._validation_base import NexsonAnnotationAdder, replace_same_agent_annotation
